@@ -2,7 +2,7 @@
 
   class.c -
 
-  $Author: nagachika $
+  $Author$
   created at: Tue Aug 10 15:05:44 JST 1993
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
@@ -25,9 +25,9 @@
 
 #include "internal.h"
 #include "ruby/st.h"
-#include "method.h"
 #include "constant.h"
 #include "vm_core.h"
+#include "id_table.h"
 #include <ctype.h>
 
 #define id_attached id__attached__
@@ -110,7 +110,7 @@ rb_class_remove_from_module_subclasses(VALUE klass)
 }
 
 void
-rb_class_foreach_subclass(VALUE klass, void(*f)(VALUE))
+rb_class_foreach_subclass(VALUE klass, void (*f)(VALUE, VALUE), VALUE arg)
 {
     rb_subclass_entry_t *cur = RCLASS_EXT(klass)->subclasses;
 
@@ -119,20 +119,32 @@ rb_class_foreach_subclass(VALUE klass, void(*f)(VALUE))
     while (cur) {
 	VALUE curklass = cur->klass;
 	cur = cur->next;
-	f(curklass);
+	f(curklass, arg);
     }
 }
 
+static void
+class_detach_subclasses(VALUE klass, VALUE arg)
+{
+    rb_class_remove_from_super_subclasses(klass);
+}
+
 void
 rb_class_detach_subclasses(VALUE klass)
 {
-    rb_class_foreach_subclass(klass, rb_class_remove_from_super_subclasses);
+    rb_class_foreach_subclass(klass, class_detach_subclasses, Qnil);
+}
+
+static void
+class_detach_module_subclasses(VALUE klass, VALUE arg)
+{
+    rb_class_remove_from_module_subclasses(klass);
 }
 
 void
 rb_class_detach_module_subclasses(VALUE klass)
 {
-    rb_class_foreach_subclass(klass, rb_class_remove_from_module_subclasses);
+    rb_class_foreach_subclass(klass, class_detach_module_subclasses, Qnil);
 }
 
 /**
@@ -151,24 +163,30 @@ static VALUE
 class_alloc(VALUE flags, VALUE klass)
 {
     NEWOBJ_OF(obj, struct RClass, klass, (flags & T_MASK) | FL_PROMOTED1 /* start from age == 2 */ | (RGENGC_WB_PROTECTED_CLASS ? FL_WB_PROTECTED : 0));
-    obj->ptr = ALLOC(rb_classext_t);
-    RCLASS_IV_TBL(obj) = 0;
-    RCLASS_CONST_TBL(obj) = 0;
-    RCLASS_M_TBL_WRAPPER(obj) = 0;
-    RCLASS_SET_SUPER((VALUE)obj, 0);
-    RCLASS_ORIGIN(obj) = (VALUE)obj;
-    RCLASS_IV_INDEX_TBL(obj) = 0;
-
-    RCLASS_EXT(obj)->subclasses = NULL;
-    RCLASS_EXT(obj)->parent_subclasses = NULL;
-    RCLASS_EXT(obj)->module_subclasses = NULL;
+    obj->ptr = ZALLOC(rb_classext_t);
+    /* ZALLOC
+      RCLASS_IV_TBL(obj) = 0;
+      RCLASS_CONST_TBL(obj) = 0;
+      RCLASS_M_TBL(obj) = 0;
+      RCLASS_IV_INDEX_TBL(obj) = 0;
+      RCLASS_SET_SUPER((VALUE)obj, 0);
+      RCLASS_EXT(obj)->subclasses = NULL;
+      RCLASS_EXT(obj)->parent_subclasses = NULL;
+      RCLASS_EXT(obj)->module_subclasses = NULL;
+     */
+    RCLASS_SET_ORIGIN((VALUE)obj, (VALUE)obj);
     RCLASS_SERIAL(obj) = rb_next_class_serial();
-
     RCLASS_REFINED_CLASS(obj) = Qnil;
     RCLASS_EXT(obj)->allocator = 0;
+
     return (VALUE)obj;
 }
 
+static void
+RCLASS_M_TBL_INIT(VALUE c)
+{
+    RCLASS_M_TBL(c) = rb_id_table_create(0);
+}
 
 /*!
  * A utility function that wraps class_alloc.
@@ -202,8 +220,8 @@ void
 rb_check_inheritable(VALUE super)
 {
     if (!RB_TYPE_P(super, T_CLASS)) {
-	rb_raise(rb_eTypeError, "superclass must be a Class (%s given)",
-		 rb_obj_classname(super));
+	rb_raise(rb_eTypeError, "superclass must be a Class (%"PRIsVALUE" given)",
+		 rb_obj_class(super));
     }
     if (RBASIC(super)->flags & FL_SINGLETON) {
 	rb_raise(rb_eTypeError, "can't make subclass of singleton class");
@@ -229,29 +247,29 @@ rb_class_new(VALUE super)
 }
 
 static void
-clone_method(VALUE klass, ID mid, const rb_method_entry_t *me)
+clone_method(VALUE old_klass, VALUE new_klass, ID mid, const rb_method_entry_t *me)
 {
-    VALUE newiseqval;
-    if (me->def && me->def->type == VM_METHOD_TYPE_ISEQ) {
-	rb_iseq_t *iseq;
-	NODE *new_cref;
-	newiseqval = rb_iseq_clone(me->def->body.iseq->self, klass);
-	GetISeqPtr(newiseqval, iseq);
-	rb_vm_rewrite_cref_stack(me->def->body.iseq->cref_stack, me->klass, klass, &new_cref);
-	RB_OBJ_WRITE(iseq->self, &iseq->cref_stack, new_cref);
-	rb_add_method(klass, mid, VM_METHOD_TYPE_ISEQ, iseq, me->flag);
-	RB_GC_GUARD(newiseqval);
+    if (me->def->type == VM_METHOD_TYPE_ISEQ) {
+	rb_cref_t *new_cref;
+	rb_vm_rewrite_cref(me->def->body.iseq.cref, old_klass, new_klass, &new_cref);
+	rb_add_method_iseq(new_klass, mid, me->def->body.iseq.iseqptr, new_cref, METHOD_ENTRY_VISI(me));
     }
     else {
-	rb_method_entry_set(klass, mid, me, me->flag);
+	rb_method_entry_set(new_klass, mid, me, METHOD_ENTRY_VISI(me));
     }
 }
 
-static int
-clone_method_i(st_data_t key, st_data_t value, st_data_t data)
+struct clone_method_arg {
+    VALUE new_klass;
+    VALUE old_klass;
+};
+
+static enum rb_id_table_iterator_result
+clone_method_i(ID key, VALUE value, void *data)
 {
-    clone_method((VALUE)data, (ID)key, (const rb_method_entry_t *)value);
-    return ST_CONTINUE;
+    const struct clone_method_arg *arg = (struct clone_method_arg *)data;
+    clone_method(arg->old_klass, arg->new_klass, key, (const rb_method_entry_t *)value);
+    return ID_TABLE_CONTINUE;
 }
 
 struct clone_const_arg {
@@ -313,10 +331,7 @@ rb_mod_init_copy(VALUE clone, VALUE orig)
 	rb_free_const_table(RCLASS_CONST_TBL(clone));
 	RCLASS_CONST_TBL(clone) = 0;
     }
-    if (RCLASS_M_TBL_WRAPPER(clone)) {
-	rb_free_m_tbl_wrapper(RCLASS_M_TBL_WRAPPER(clone));
-	RCLASS_M_TBL_WRAPPER(clone) = 0;
-    }
+    RCLASS_M_TBL(clone) = 0;
     if (RCLASS_IV_TBL(orig)) {
 	st_data_t id;
 
@@ -337,8 +352,11 @@ rb_mod_init_copy(VALUE clone, VALUE orig)
 	st_foreach(RCLASS_CONST_TBL(orig), clone_const_i, (st_data_t)&arg);
     }
     if (RCLASS_M_TBL(orig)) {
+	struct clone_method_arg arg;
+	arg.old_klass = orig;
+	arg.new_klass = clone;
 	RCLASS_M_TBL_INIT(clone);
-	st_foreach(RCLASS_M_TBL(orig), clone_method_i, (st_data_t)clone);
+	rb_id_table_foreach(RCLASS_M_TBL(orig), clone_method_i, &arg);
     }
 
     return clone;
@@ -353,7 +371,7 @@ rb_singleton_class_clone(VALUE obj)
 VALUE
 rb_singleton_class_clone_and_attach(VALUE obj, VALUE attach)
 {
-    VALUE klass = RBASIC(obj)->klass;
+    const VALUE klass = RBASIC(obj)->klass;
 
     if (!FL_TEST(klass, FL_SINGLETON))
 	return klass;
@@ -384,7 +402,12 @@ rb_singleton_class_clone_and_attach(VALUE obj, VALUE attach)
 	    rb_singleton_class_attached(clone, attach);
 	}
 	RCLASS_M_TBL_INIT(clone);
-	st_foreach(RCLASS_M_TBL(klass), clone_method_i, (st_data_t)clone);
+	{
+	    struct clone_method_arg arg;
+	    arg.old_klass = klass;
+	    arg.new_klass = clone;
+	    rb_id_table_foreach(RCLASS_M_TBL(klass), clone_method_i, &arg);
+	}
 	rb_singleton_class_attached(RBASIC(clone)->klass, clone);
 	FL_SET(clone, FL_SINGLETON);
 
@@ -403,7 +426,7 @@ rb_singleton_class_attached(VALUE klass, VALUE obj)
 	if (!RCLASS_IV_TBL(klass)) {
 	    RCLASS_IV_TBL(klass) = st_init_numtable();
 	}
-	rb_st_insert_id_and_value(klass, RCLASS_IV_TBL(klass), id_attached, obj);
+	rb_class_ivar_set(klass, id_attached, obj);
     }
 }
 
@@ -419,6 +442,19 @@ rb_singleton_class_attached(VALUE klass, VALUE obj)
  */
 #define META_CLASS_OF_CLASS_CLASS_P(k)  (METACLASS_OF(k) == (k))
 
+static int
+rb_singleton_class_has_metaclass_p(VALUE sklass)
+{
+    return rb_attr_get(METACLASS_OF(sklass), id_attached) == sklass;
+}
+
+int
+rb_singleton_class_internal_p(VALUE sklass)
+{
+    return (RB_TYPE_P(rb_attr_get(sklass, id_attached), T_CLASS) &&
+	    !rb_singleton_class_has_metaclass_p(sklass));
+}
+
 /*!
  * whether k has a metaclass
  * @retval 1 if \a k has a metaclass
@@ -426,7 +462,7 @@ rb_singleton_class_attached(VALUE klass, VALUE obj)
  */
 #define HAVE_METACLASS_P(k) \
     (FL_TEST(METACLASS_OF(k), FL_SINGLETON) && \
-     rb_ivar_get(METACLASS_OF(k), id_attached) == (k))
+     rb_singleton_class_has_metaclass_p(k))
 
 /*!
  * ensures \a klass belongs to its own eigenclass.
@@ -513,6 +549,7 @@ Init_class_hierarchy(void)
 {
     rb_cBasicObject = boot_defclass("BasicObject", 0);
     rb_cObject = boot_defclass("Object", rb_cBasicObject);
+    rb_gc_register_mark_object(rb_cObject);
     rb_cModule = boot_defclass("Module", rb_cObject);
     rb_cClass =  boot_defclass("Class",  rb_cModule);
 
@@ -613,7 +650,8 @@ rb_define_class(const char *name, VALUE super)
     if (rb_const_defined(rb_cObject, id)) {
 	klass = rb_const_get(rb_cObject, id);
 	if (!RB_TYPE_P(klass, T_CLASS)) {
-	    rb_raise(rb_eTypeError, "%s is not a class", name);
+	    rb_raise(rb_eTypeError, "%s is not a class (%"PRIsVALUE")",
+		     name, rb_obj_class(klass));
 	}
 	if (rb_class_real(RCLASS_SUPER(klass)) != super) {
 	    rb_raise(rb_eTypeError, "superclass mismatch for class %s", name);
@@ -642,7 +680,7 @@ rb_define_class(const char *name, VALUE super)
  * \return the created class
  * \throw TypeError if the constant name \a name is already taken but
  *                  the constant is not a \c Class.
- * \throw NameError if the class is already defined but the class can not
+ * \throw TypeError if the class is already defined but the class can not
  *                  be reopened because its superclass is not \a super.
  * \post top-level constant named \a name refers the returned class.
  *
@@ -665,7 +703,7 @@ rb_define_class_under(VALUE outer, const char *name, VALUE super)
  * \return the created class
  * \throw TypeError if the constant name \a name is already taken but
  *                  the constant is not a \c Class.
- * \throw NameError if the class is already defined but the class can not
+ * \throw TypeError if the class is already defined but the class can not
  *                  be reopened because its superclass is not \a super.
  * \post top-level constant named \a name refers the returned class.
  *
@@ -680,10 +718,15 @@ rb_define_class_id_under(VALUE outer, ID id, VALUE super)
     if (rb_const_defined_at(outer, id)) {
 	klass = rb_const_get_at(outer, id);
 	if (!RB_TYPE_P(klass, T_CLASS)) {
-	    rb_raise(rb_eTypeError, "%"PRIsVALUE" is not a class", rb_id2str(id));
+	    rb_raise(rb_eTypeError, "%"PRIsVALUE"::%"PRIsVALUE" is not a class"
+		     " (%"PRIsVALUE")",
+		     outer, rb_id2str(id), rb_obj_class(klass));
 	}
 	if (rb_class_real(RCLASS_SUPER(klass)) != super) {
-	    rb_name_error(id, "%"PRIsVALUE" is already defined", rb_id2str(id));
+	    rb_raise(rb_eTypeError, "superclass mismatch for class "
+		     "%"PRIsVALUE"::%"PRIsVALUE""
+		     " (%"PRIsVALUE" is given but was %"PRIsVALUE")",
+		     outer, rb_id2str(id), RCLASS_SUPER(klass), super);
 	}
 	return klass;
     }
@@ -728,9 +771,11 @@ rb_define_module(const char *name)
     id = rb_intern(name);
     if (rb_const_defined(rb_cObject, id)) {
 	module = rb_const_get(rb_cObject, id);
-	if (RB_TYPE_P(module, T_MODULE))
-	    return module;
-	rb_raise(rb_eTypeError, "%s is not a module", rb_obj_classname(module));
+	if (!RB_TYPE_P(module, T_MODULE)) {
+	    rb_raise(rb_eTypeError, "%s is not a module (%"PRIsVALUE")",
+		     name, rb_obj_class(module));
+	}
+	return module;
     }
     module = rb_define_module_id(id);
     rb_vm_add_root_module(id, module);
@@ -752,10 +797,12 @@ rb_define_module_id_under(VALUE outer, ID id)
 
     if (rb_const_defined_at(outer, id)) {
 	module = rb_const_get_at(outer, id);
-	if (RB_TYPE_P(module, T_MODULE))
-	    return module;
-	rb_raise(rb_eTypeError, "%s::%s is not a module",
-		 rb_class2name(outer), rb_obj_classname(module));
+	if (!RB_TYPE_P(module, T_MODULE)) {
+	    rb_raise(rb_eTypeError, "%"PRIsVALUE"::%"PRIsVALUE" is not a module"
+		     " (%"PRIsVALUE")",
+		     outer, rb_id2str(id), rb_obj_class(module));
+	}
+	return module;
     }
     module = rb_define_module_id(id);
     rb_const_set(outer, id, module);
@@ -782,8 +829,8 @@ rb_include_class_new(VALUE module, VALUE super)
     RCLASS_IV_TBL(klass) = RCLASS_IV_TBL(module);
     RCLASS_CONST_TBL(klass) = RCLASS_CONST_TBL(module);
 
-    RCLASS_M_TBL_WRAPPER(OBJ_WB_UNPROTECT(klass)) =
-	RCLASS_M_TBL_WRAPPER(OBJ_WB_UNPROTECT(RCLASS_ORIGIN(module)));
+    RCLASS_M_TBL(OBJ_WB_UNPROTECT(klass)) =
+      RCLASS_M_TBL(OBJ_WB_UNPROTECT(RCLASS_ORIGIN(module))); /* TODO: unprotected? */
 
     RCLASS_SET_SUPER(klass, super);
     if (RB_TYPE_P(module, T_ICLASS)) {
@@ -798,39 +845,44 @@ rb_include_class_new(VALUE module, VALUE super)
     return (VALUE)klass;
 }
 
-static int include_modules_at(const VALUE klass, VALUE c, VALUE module);
+static int include_modules_at(const VALUE klass, VALUE c, VALUE module, int search_super);
+
+static void
+ensure_includable(VALUE klass, VALUE module)
+{
+    rb_frozen_class_p(klass);
+    Check_Type(module, T_MODULE);
+    if (!NIL_P(rb_refinement_module_get_refined_class(module))) {
+	rb_raise(rb_eArgError, "refinement module is not allowed");
+    }
+    OBJ_INFECT(klass, module);
+}
 
 void
 rb_include_module(VALUE klass, VALUE module)
 {
     int changed = 0;
 
-    rb_frozen_class_p(klass);
+    ensure_includable(klass, module);
 
-    if (!RB_TYPE_P(module, T_MODULE)) {
-	Check_Type(module, T_MODULE);
-    }
-
-    OBJ_INFECT(klass, module);
-
-    changed = include_modules_at(klass, RCLASS_ORIGIN(klass), module);
+    changed = include_modules_at(klass, RCLASS_ORIGIN(klass), module, TRUE);
     if (changed < 0)
 	rb_raise(rb_eArgError, "cyclic include detected");
 }
 
-static int
-add_refined_method_entry_i(st_data_t key, st_data_t value, st_data_t data)
+static enum rb_id_table_iterator_result
+add_refined_method_entry_i(ID key, VALUE value, void *data)
 {
-    rb_add_refined_method_entry((VALUE) data, (ID) key);
-    return ST_CONTINUE;
+    rb_add_refined_method_entry((VALUE)data, key);
+    return ID_TABLE_CONTINUE;
 }
 
 static int
-include_modules_at(const VALUE klass, VALUE c, VALUE module)
+include_modules_at(const VALUE klass, VALUE c, VALUE module, int search_super)
 {
     VALUE p, iclass;
     int method_changed = 0, constant_changed = 0;
-    const st_table *const klass_m_tbl = RCLASS_M_TBL(RCLASS_ORIGIN(klass));
+    struct rb_id_table *const klass_m_tbl = RCLASS_M_TBL(RCLASS_ORIGIN(klass));
 
     while (module) {
 	int superclass_seen = FALSE;
@@ -841,42 +893,38 @@ include_modules_at(const VALUE klass, VALUE c, VALUE module)
 	    return -1;
 	/* ignore if the module included already in superclasses */
 	for (p = RCLASS_SUPER(klass); p; p = RCLASS_SUPER(p)) {
-	    switch (BUILTIN_TYPE(p)) {
-	      case T_ICLASS:
-		if (RCLASS_M_TBL_WRAPPER(p) == RCLASS_M_TBL_WRAPPER(module)) {
+	    int type = BUILTIN_TYPE(p);
+	    if (type == T_ICLASS) {
+		if (RCLASS_M_TBL(p) == RCLASS_M_TBL(module)) {
 		    if (!superclass_seen) {
 			c = p;  /* move insertion point */
 		    }
 		    goto skip;
 		}
-		break;
-	      case T_CLASS:
+	    }
+	    else if (type == T_CLASS) {
+		if (!search_super) break;
 		superclass_seen = TRUE;
-		break;
 	    }
 	}
 	iclass = rb_include_class_new(module, RCLASS_SUPER(c));
 	c = RCLASS_SET_SUPER(c, iclass);
 
-	if (BUILTIN_TYPE(module) == T_ICLASS) {
-	    rb_module_add_to_subclasses_list(RBASIC(module)->klass, iclass);
-	}
-	else {
-	    rb_module_add_to_subclasses_list(module, iclass);
+	{
+	    VALUE m = module;
+	    if (BUILTIN_TYPE(m) == T_ICLASS) m = RBASIC(m)->klass;
+	    rb_module_add_to_subclasses_list(m, iclass);
 	}
 
 	if (FL_TEST(klass, RMODULE_IS_REFINEMENT)) {
 	    VALUE refined_class =
 		rb_refinement_module_get_refined_class(klass);
 
-	    st_foreach(RMODULE_M_TBL(module), add_refined_method_entry_i,
-		       (st_data_t) refined_class);
+	    rb_id_table_foreach(RMODULE_M_TBL(module), add_refined_method_entry_i, (void *)refined_class);
 	    FL_SET(c, RMODULE_INCLUDED_INTO_REFINEMENT);
 	}
-	if (RMODULE_M_TBL(module) && RMODULE_M_TBL(module)->num_entries)
-	    method_changed = 1;
-	if (RMODULE_CONST_TBL(module) && RMODULE_CONST_TBL(module)->num_entries)
-	    constant_changed = 1;
+	if (RMODULE_M_TBL(module) && rb_id_table_size(RMODULE_M_TBL(module))) method_changed = 1;
+	if (RMODULE_CONST_TBL(module) && RMODULE_CONST_TBL(module)->num_entries) constant_changed = 1;
       skip:
 	module = RCLASS_SUPER(module);
     }
@@ -887,30 +935,30 @@ include_modules_at(const VALUE klass, VALUE c, VALUE module)
     return method_changed;
 }
 
-static int
-move_refined_method(st_data_t key, st_data_t value, st_data_t data)
+static enum rb_id_table_iterator_result
+move_refined_method(ID key, VALUE value, void *data)
 {
     rb_method_entry_t *me = (rb_method_entry_t *) value;
-    st_table *tbl = (st_table *) data;
+    VALUE klass = (VALUE)data;
+    struct rb_id_table *tbl = RCLASS_M_TBL(klass);
 
     if (me->def->type == VM_METHOD_TYPE_REFINED) {
-	if (me->def->body.orig_me) {
-	    rb_method_entry_t *orig_me = me->def->body.orig_me, *new_me;
-	    me->def->body.orig_me = NULL;
-	    new_me = ALLOC(rb_method_entry_t);
-	    *new_me = *me;
-	    st_add_direct(tbl, key, (st_data_t) new_me);
-	    *me = *orig_me;
-	    xfree(orig_me);
-	    return ST_CONTINUE;
+	if (me->def->body.refined.orig_me) {
+	    const rb_method_entry_t *orig_me = me->def->body.refined.orig_me, *new_me;
+	    RB_OBJ_WRITE(me, &me->def->body.refined.orig_me, NULL);
+	    new_me = rb_method_entry_clone(me);
+	    rb_id_table_insert(tbl, key, (VALUE)new_me);
+	    RB_OBJ_WRITTEN(klass, Qundef, new_me);
+	    rb_method_entry_copy(me, orig_me);
+	    return ID_TABLE_CONTINUE;
 	}
 	else {
-	    st_add_direct(tbl, key, (st_data_t) me);
-	    return ST_DELETE;
+	    rb_id_table_insert(tbl, key, (VALUE)me);
+	    return ID_TABLE_DELETE;
 	}
     }
     else {
-	return ST_CONTINUE;
+	return ID_TABLE_CONTINUE;
     }
 }
 
@@ -920,11 +968,7 @@ rb_prepend_module(VALUE klass, VALUE module)
     VALUE origin;
     int changed = 0;
 
-    rb_frozen_class_p(klass);
-
-    Check_Type(module, T_MODULE);
-
-    OBJ_INFECT(klass, module);
+    ensure_includable(klass, module);
 
     origin = RCLASS_ORIGIN(klass);
     if (origin == klass) {
@@ -932,13 +976,12 @@ rb_prepend_module(VALUE klass, VALUE module)
 	OBJ_WB_UNPROTECT(origin); /* TODO: conservative shading. Need more survey. */
 	RCLASS_SET_SUPER(origin, RCLASS_SUPER(klass));
 	RCLASS_SET_SUPER(klass, origin);
-	RB_OBJ_WRITE(klass, &RCLASS_ORIGIN(klass), origin);
-	RCLASS_M_TBL_WRAPPER(origin) = RCLASS_M_TBL_WRAPPER(klass);
+	RCLASS_SET_ORIGIN(klass, origin);
+	RCLASS_M_TBL(origin) = RCLASS_M_TBL(klass);
 	RCLASS_M_TBL_INIT(klass);
-	st_foreach(RCLASS_M_TBL(origin), move_refined_method,
-		   (st_data_t) RCLASS_M_TBL(klass));
+	rb_id_table_foreach(RCLASS_M_TBL(origin), move_refined_method, (void *)klass);
     }
-    changed = include_modules_at(klass, klass, module);
+    changed = include_modules_at(klass, klass, module, FALSE);
     if (changed < 0)
 	rb_raise(rb_eArgError, "cyclic prepend detected");
     if (changed) {
@@ -1047,52 +1090,46 @@ rb_mod_ancestors(VALUE mod)
     return ary;
 }
 
-#define VISI(x) ((x)&NOEX_MASK)
-#define VISI_CHECK(x,f) (VISI(x) == (f))
-
 static int
-ins_methods_push(ID name, long type, VALUE ary, long visi)
+ins_methods_push(ID name, rb_method_visibility_t visi, VALUE ary, rb_method_visibility_t expected_visi)
 {
-    if (type == -1) return ST_CONTINUE;
+    if (visi == METHOD_VISI_UNDEF) return ST_CONTINUE;
 
-    switch (visi) {
-      case NOEX_PRIVATE:
-      case NOEX_PROTECTED:
-      case NOEX_PUBLIC:
-	visi = (type == visi);
+    switch (expected_visi) {
+      case METHOD_VISI_UNDEF:
+	if (visi != METHOD_VISI_PRIVATE) rb_ary_push(ary, ID2SYM(name));
 	break;
-      default:
-	visi = (type != NOEX_PRIVATE);
+      case METHOD_VISI_PRIVATE:
+      case METHOD_VISI_PROTECTED:
+      case METHOD_VISI_PUBLIC:
+	if (visi == expected_visi) rb_ary_push(ary, ID2SYM(name));
 	break;
     }
-    if (visi) {
-	rb_ary_push(ary, ID2SYM(name));
-    }
     return ST_CONTINUE;
 }
 
 static int
 ins_methods_i(st_data_t name, st_data_t type, st_data_t ary)
 {
-    return ins_methods_push((ID)name, (long)type, (VALUE)ary, -1); /* everything but private */
+    return ins_methods_push((ID)name, (rb_method_visibility_t)type, (VALUE)ary, METHOD_VISI_UNDEF); /* everything but private */
 }
 
 static int
 ins_methods_prot_i(st_data_t name, st_data_t type, st_data_t ary)
 {
-    return ins_methods_push((ID)name, (long)type, (VALUE)ary, NOEX_PROTECTED);
+    return ins_methods_push((ID)name, (rb_method_visibility_t)type, (VALUE)ary, METHOD_VISI_PROTECTED);
 }
 
 static int
 ins_methods_priv_i(st_data_t name, st_data_t type, st_data_t ary)
 {
-    return ins_methods_push((ID)name, (long)type, (VALUE)ary, NOEX_PRIVATE);
+    return ins_methods_push((ID)name, (rb_method_visibility_t)type, (VALUE)ary, METHOD_VISI_PRIVATE);
 }
 
 static int
 ins_methods_pub_i(st_data_t name, st_data_t type, st_data_t ary)
 {
-    return ins_methods_push((ID)name, (long)type, (VALUE)ary, NOEX_PUBLIC);
+    return ins_methods_push((ID)name, (rb_method_visibility_t)type, (VALUE)ary, METHOD_VISI_PUBLIC);
 }
 
 struct method_entry_arg {
@@ -1100,29 +1137,29 @@ struct method_entry_arg {
     int recur;
 };
 
-static int
-method_entry_i(st_data_t key, st_data_t value, st_data_t data)
+static enum rb_id_table_iterator_result
+method_entry_i(ID key, VALUE value, void *data)
 {
     const rb_method_entry_t *me = (const rb_method_entry_t *)value;
     struct method_entry_arg *arg = (struct method_entry_arg *)data;
-    long type;
+    rb_method_visibility_t type;
 
-    if (me && me->def->type == VM_METHOD_TYPE_REFINED) {
-	VALUE klass = me->klass;
-	me = rb_resolve_refined_method(Qnil, me, NULL);
-	if (!me) return ST_CONTINUE;
-	if (!arg->recur && me->klass != klass) return ST_CONTINUE;
+    if (me->def->type == VM_METHOD_TYPE_REFINED) {
+	VALUE owner = me->owner;
+	me = rb_resolve_refined_method(Qnil, me);
+	if (!me) return ID_TABLE_CONTINUE;
+	if (!arg->recur && me->owner != owner) return ID_TABLE_CONTINUE;
     }
     if (!st_lookup(arg->list, key, 0)) {
 	if (UNDEFINED_METHOD_ENTRY_P(me)) {
-	    type = -1; /* none */
+	    type = METHOD_VISI_UNDEF; /* none */
 	}
 	else {
-	    type = VISI(me->flag);
+	    type = METHOD_ENTRY_VISI(me);
 	}
-	st_add_direct(arg->list, key, type);
+	st_add_direct(arg->list, key, (st_data_t)type);
     }
-    return ST_CONTINUE;
+    return ID_TABLE_CONTINUE;
 }
 
 static VALUE
@@ -1149,7 +1186,7 @@ class_instance_method_list(int argc, const VALUE *argv, VALUE mod, int obj, int
     me_arg.list = st_init_numtable();
     me_arg.recur = recur;
     for (; mod; mod = RCLASS_SUPER(mod)) {
-	if (RCLASS_M_TBL(mod)) st_foreach(RCLASS_M_TBL(mod), method_entry_i, (st_data_t)&me_arg);
+	if (RCLASS_M_TBL(mod)) rb_id_table_foreach(RCLASS_M_TBL(mod), method_entry_i, &me_arg);
 	if (BUILTIN_TYPE(mod) == T_ICLASS && !prepended) continue;
 	if (obj && FL_TEST(mod, FL_SINGLETON)) continue;
 	if (!recur) break;
@@ -1370,7 +1407,7 @@ rb_obj_singleton_methods(int argc, const VALUE *argv, VALUE obj)
 {
     VALUE recur, ary, klass, origin;
     struct method_entry_arg me_arg;
-    st_table *mtbl;
+    struct rb_id_table *mtbl;
 
     if (argc == 0) {
 	recur = Qtrue;
@@ -1383,14 +1420,12 @@ rb_obj_singleton_methods(int argc, const VALUE *argv, VALUE obj)
     me_arg.list = st_init_numtable();
     me_arg.recur = RTEST(recur);
     if (klass && FL_TEST(klass, FL_SINGLETON)) {
-	if ((mtbl = RCLASS_M_TBL(origin)) != 0)
-	    st_foreach(mtbl, method_entry_i, (st_data_t)&me_arg);
+	if ((mtbl = RCLASS_M_TBL(origin)) != 0) rb_id_table_foreach(mtbl, method_entry_i, &me_arg);
 	klass = RCLASS_SUPER(klass);
     }
     if (RTEST(recur)) {
 	while (klass && (FL_TEST(klass, FL_SINGLETON) || RB_TYPE_P(klass, T_ICLASS))) {
-	    if (klass != origin && (mtbl = RCLASS_M_TBL(klass)) != 0)
-		st_foreach(mtbl, method_entry_i, (st_data_t)&me_arg);
+	    if (klass != origin && (mtbl = RCLASS_M_TBL(klass)) != 0) rb_id_table_foreach(mtbl, method_entry_i, &me_arg);
 	    klass = RCLASS_SUPER(klass);
 	}
     }
@@ -1461,31 +1496,31 @@ rb_obj_singleton_methods(int argc, const VALUE *argv, VALUE obj)
 void
 rb_define_method_id(VALUE klass, ID mid, VALUE (*func)(ANYARGS), int argc)
 {
-    rb_add_method_cfunc(klass, mid, func, argc, NOEX_PUBLIC);
+    rb_add_method_cfunc(klass, mid, func, argc, METHOD_VISI_PUBLIC);
 }
 
 void
 rb_define_method(VALUE klass, const char *name, VALUE (*func)(ANYARGS), int argc)
 {
-    rb_add_method_cfunc(klass, rb_intern(name), func, argc, NOEX_PUBLIC);
+    rb_add_method_cfunc(klass, rb_intern(name), func, argc, METHOD_VISI_PUBLIC);
 }
 
 void
 rb_define_protected_method(VALUE klass, const char *name, VALUE (*func)(ANYARGS), int argc)
 {
-    rb_add_method_cfunc(klass, rb_intern(name), func, argc, NOEX_PROTECTED);
+    rb_add_method_cfunc(klass, rb_intern(name), func, argc, METHOD_VISI_PROTECTED);
 }
 
 void
 rb_define_private_method(VALUE klass, const char *name, VALUE (*func)(ANYARGS), int argc)
 {
-    rb_add_method_cfunc(klass, rb_intern(name), func, argc, NOEX_PRIVATE);
+    rb_add_method_cfunc(klass, rb_intern(name), func, argc, METHOD_VISI_PRIVATE);
 }
 
 void
 rb_undef_method(VALUE klass, const char *name)
 {
-    rb_add_method(klass, rb_intern(name), VM_METHOD_TYPE_UNDEF, 0, NOEX_UNDEF);
+    rb_add_method(klass, rb_intern(name), VM_METHOD_TYPE_UNDEF, 0, METHOD_VISI_UNDEF);
 }
 
 /*!
@@ -1531,7 +1566,8 @@ singleton_class_of(VALUE obj)
 {
     VALUE klass;
 
-    if (FIXNUM_P(obj) || FLONUM_P(obj) || SYMBOL_P(obj)) {
+    if (FIXNUM_P(obj) || FLONUM_P(obj) || STATIC_SYM_P(obj)) {
+      no_singleton:
 	rb_raise(rb_eTypeError, "can't define singleton");
     }
     if (SPECIAL_CONST_P(obj)) {
@@ -1541,9 +1577,12 @@ singleton_class_of(VALUE obj)
 	return klass;
     }
     else {
-	enum ruby_value_type type = BUILTIN_TYPE(obj);
-	if (type == T_FLOAT || type == T_BIGNUM) {
-	    rb_raise(rb_eTypeError, "can't define singleton");
+	switch (BUILTIN_TYPE(obj)) {
+	  case T_FLOAT: case T_BIGNUM: case T_SYMBOL:
+	    goto no_singleton;
+	  case T_STRING:
+	    if (FL_TEST_RAW(obj, RSTRING_FSTR)) goto no_singleton;
+	    break;
 	}
     }
 
@@ -1706,7 +1745,7 @@ rb_define_attr(VALUE klass, const char *name, int read, int write)
 int
 rb_obj_basic_to_s_p(VALUE obj)
 {
-    const rb_method_entry_t *me = rb_method_entry(CLASS_OF(obj), rb_intern("to_s"), 0);
+    const rb_method_entry_t *me = rb_method_entry(CLASS_OF(obj), rb_intern("to_s"));
     if (me && me->def && me->def->type == VM_METHOD_TYPE_CFUNC &&
 	me->def->body.cfunc.func == rb_any_to_s)
 	return 1;
@@ -1885,7 +1924,7 @@ unknown_keyword_error(VALUE hash, const ID *table, int keywords)
 	st_data_t key = ID2SYM(table[i]);
 	st_delete(tbl, &key, NULL);
     }
-    keys = rb_funcall(hash, rb_intern("keys"), 0, 0);
+    keys = rb_funcallv(hash, rb_intern("keys"), 0, 0);
     if (!RB_TYPE_P(keys, T_ARRAY)) rb_raise(rb_eArgError, "unknown keyword");
     rb_keyword_error("unknown", keys);
 }
@@ -1913,6 +1952,9 @@ rb_extract_keywords(VALUE *orighash)
     }
     st_foreach(rb_hash_tbl_raw(hash), separate_symbol, (st_data_t)&parthash);
     *orighash = parthash[1];
+    if (parthash[1] && RBASIC_CLASS(hash) != rb_cHash) {
+	RBASIC_SET_CLASS(parthash[1], RBASIC_CLASS(hash));
+    }
     return parthash[0];
 }
 
@@ -1968,7 +2010,7 @@ rb_get_kwargs(VALUE keyword_hash, const ID *table, int required, int optional, V
 	}
     }
     if (!rest && keyword_hash) {
-	if (RHASH_SIZE(keyword_hash) > (unsigned int)j) {
+	if (RHASH_SIZE(keyword_hash) > (unsigned int)(values ? 0 : j)) {
 	    unknown_keyword_error(keyword_hash, table, required+optional);
 	}
     }
@@ -1976,6 +2018,12 @@ rb_get_kwargs(VALUE keyword_hash, const ID *table, int required, int optional, V
 #undef extract_kwarg
 }
 
+int
+rb_class_has_methods(VALUE c)
+{
+    return rb_id_table_size(RCLASS_M_TBL(c)) == 0 ? FALSE : TRUE;
+}
+
 /*!
  * \}
  */
