@@ -231,7 +231,7 @@ code_to_mbclen(OnigCodePoint code, OnigEncoding enc ARG_UNUSED)
   }
   else if (code <= 0xffff) {
     int low = code & 0xff;
-    if (low < 0x40 || low == 0x7f || 0xfc < low)
+    if (! SJIS_ISMB_TRAIL(low))
       return ONIGERR_INVALID_CODE_POINT_VALUE;
     return 2;
   }
@@ -278,7 +278,7 @@ apply_all_case_fold(OnigCaseFoldType flag,
 		    OnigApplyAllCaseFoldFunc f, void* arg, OnigEncoding enc)
 {
   return onigenc_apply_all_case_fold_with_map(
-            sizeof(CaseFoldMap)/sizeof(OnigPairCaseFoldCodes), CaseFoldMap, 0,
+            numberof(CaseFoldMap), CaseFoldMap, 0,
             flag, f, arg);
 }
 
@@ -433,12 +433,6 @@ is_allowed_reverse_match(const UChar* s, const UChar* end, OnigEncoding enc ARG_
 }
 
 
-static int PropertyInited = 0;
-static const OnigCodePoint** PropertyList;
-static int PropertyListNum;
-static int PropertyListSize;
-static hash_table_type* PropertyNameTable;
-
 static const OnigCodePoint CR_Hiragana[] = {
   1,
   0x829f, 0x82f1
@@ -493,41 +487,20 @@ static const OnigCodePoint CR_Cyrillic[] = {
   0x8480, 0x8491,
 }; /* CR_Cyrillic */
 
-static int
-init_property_list(void)
-{
-  int r;
-
-  PROPERTY_LIST_ADD_PROP("hiragana", CR_Hiragana);
-  PROPERTY_LIST_ADD_PROP("katakana", CR_Katakana);
-  PROPERTY_LIST_ADD_PROP("han", CR_Han);
-  PROPERTY_LIST_ADD_PROP("latin", CR_Latin);
-  PROPERTY_LIST_ADD_PROP("greek", CR_Greek);
-  PROPERTY_LIST_ADD_PROP("cyrillic", CR_Cyrillic);
-  PropertyInited = 1;
-
- end:
-  return r;
-}
+#include "enc/jis/props.h"
 
 static int
-property_name_to_ctype(OnigEncoding enc, UChar* p, UChar* end)
+property_name_to_ctype(OnigEncoding enc, const UChar* p, const UChar* end)
 {
-  hash_data_type ctype;
-  UChar *s, *e;
+  const UChar *s = p, *e = end;
+  const struct enc_property *prop =
+    onig_jis_property((const char* )s, (unsigned int )(e - s));
 
-  PROPERTY_LIST_INIT_CHECK;
-
-  s = e = ALLOCA_N(UChar, end-p+1);
-  for (; p < end; p++) {
-    *e++ = ONIGENC_ASCII_CODE_TO_LOWER_CASE(*p);
-  }
-
-  if (onig_st_lookup_strend(PropertyNameTable, s, e, &ctype) == 0) {
+  if (!prop) {
     return onigenc_minimum_property_name_to_ctype(enc, s, e);
   }
 
-  return (int )ctype;
+  return (int )prop->ctype;
 }
 
 static int
@@ -543,8 +516,6 @@ is_code_ctype(OnigCodePoint code, unsigned int ctype, OnigEncoding enc)
     }
   }
   else {
-    PROPERTY_LIST_INIT_CHECK;
-
     ctype -= (ONIGENC_MAX_STD_CTYPE + 1);
     if (ctype >= (unsigned int )PropertyListNum)
       return ONIGERR_TYPE_BUG;
@@ -565,8 +536,6 @@ get_ctype_code_range(OnigCtype ctype, OnigCodePoint* sb_out,
   else {
     *sb_out = 0x80;
 
-    PROPERTY_LIST_INIT_CHECK;
-
     ctype -= (ONIGENC_MAX_STD_CTYPE + 1);
     if (ctype >= (OnigCtype )PropertyListNum)
       return ONIGERR_TYPE_BUG;
