@@ -1,5 +1,6 @@
+# frozen_string_literal: false
 #
-# $Id: core.rb 11708 2007-02-12 23:01:19Z shyouhei $
+# $Id$
 #
 # Copyright (c) 2003-2005 Minero Aoki
 #
@@ -12,8 +13,8 @@ require 'ripper.so'
 
 class Ripper
 
-  # Parses Ruby program read from _src_.
-  # _src_ must be a String or a IO or a object which has #gets method.
+  # Parses the given Ruby program read from +src+.
+  # +src+ must be a String or an IO or a object with a #gets method.
   def Ripper.parse(src, filename = '(ripper)', lineno = 1)
     new(src, filename, lineno).parse
   end
@@ -29,25 +30,30 @@ class Ripper
 
   private
 
+  def _dispatch_0() nil end
+  def _dispatch_1(a) a end
+  def _dispatch_2(a, b) a end
+  def _dispatch_3(a, b, c) a end
+  def _dispatch_4(a, b, c, d) a end
+  def _dispatch_5(a, b, c, d, e) a end
+  def _dispatch_6(a, b, c, d, e, f) a end
+  def _dispatch_7(a, b, c, d, e, f, g) a end
+
   #
   # Parser Events
   #
 
   PARSER_EVENT_TABLE.each do |id, arity|
-    module_eval(<<-End, __FILE__, __LINE__ + 1)
-      def on_#{id}(#{ ('a'..'z').to_a[0, arity].join(', ') })
-        #{arity == 0 ? 'nil' : 'a'}
-      end
-    End
+    alias_method "on_#{id}", "_dispatch_#{arity}"
   end
 
   # This method is called when weak warning is produced by the parser.
-  # _fmt_ and _args_ is printf style.
+  # +fmt+ and +args+ is printf style.
   def warn(fmt, *args)
   end
 
   # This method is called when strong warning is produced by the parser.
-  # _fmt_ and _args_ is printf style.
+  # +fmt+ and +args+ is printf style.
   def warning(fmt, *args)
   end
 
@@ -60,11 +66,7 @@ class Ripper
   #
 
   SCANNER_EVENTS.each do |id|
-    module_eval(<<-End, __FILE__, __LINE__ + 1)
-      def on_#{id}(token)
-        token
-      end
-    End
+    alias_method "on_#{id}", :_dispatch_1
   end
 
 end
