@@ -1,5 +1,6 @@
+# frozen_string_literal: false
 #
-# $Id: lexer.rb 13966 2007-11-19 07:10:09Z matz $
+# $Id$
 #
 # Copyright (c) 2004,2005 Minero Aoki
 #
@@ -12,18 +13,22 @@ require 'ripper/core'
 
 class Ripper
 
-  # Tokenizes Ruby program and returns an Array of String.
+  # Tokenizes the Ruby program and returns an array of strings.
+  #
+  #   p Ripper.tokenize("def m(a) nil end")
+  #      # => ["def", " ", "m", "(", "a", ")", " ", "nil", " ", "end"]
+  #
   def Ripper.tokenize(src, filename = '-', lineno = 1)
     Lexer.new(src, filename, lineno).tokenize
   end
 
-  # Tokenizes Ruby program and returns an Array of Array,
-  # which is formatted like [[lineno, column], type, token].
+  # Tokenizes the Ruby program and returns an array of an array,
+  # which is formatted like <code>[[lineno, column], type, token]</code>.
   #
   #   require 'ripper'
   #   require 'pp'
   #
-  #   p Ripper.lex("def m(a) nil end")
+  #   pp Ripper.lex("def m(a) nil end")
   #     #=> [[[1,  0], :on_kw,     "def"],
   #          [[1,  3], :on_sp,     " "  ],
   #          [[1,  4], :on_ident,  "m"  ],
@@ -40,28 +45,63 @@ class Ripper
   end
 
   class Lexer < ::Ripper   #:nodoc: internal use only
+    Elem = Struct.new(:pos, :event, :tok)
+
     def tokenize
-      lex().map {|pos, event, tok| tok }
+      parse().sort_by(&:pos).map(&:tok)
     end
 
     def lex
-      parse().sort_by {|pos, event, tok| pos }
+      parse().sort_by(&:pos).map(&:to_a)
     end
 
     def parse
       @buf = []
+      @stack = []
       super
+      @buf.flatten!
       @buf
     end
 
     private
 
-    SCANNER_EVENTS.each do |event|
-      module_eval(<<-End, __FILE__+'/module_eval', __LINE__ + 1)
-        def on_#{event}(tok)
-          @buf.push [[lineno(), column()], :on_#{event}, tok]
+    def on_heredoc_dedent(v, w)
+      ignored_sp = []
+      heredoc = @buf.last
+      heredoc.each_with_index do |e, i|
+        if Elem === e and e.event == :on_tstring_content
+          tok = e.tok.dup if w > 0 and /\A\s/ =~ e.tok
+          if (n = dedent_string(e.tok, w)) > 0
+            ignored_sp << [i, Elem.new(e.pos.dup, :on_ignored_sp, tok[0, n])]
+            e.pos[1] += n
+          end
         end
-      End
+      end
+      ignored_sp.reverse_each do |i, e|
+        heredoc[i, 0] = [e]
+      end
+      v
+    end
+
+    def on_heredoc_beg(tok)
+      @stack.push @buf
+      buf = []
+      @buf << buf
+      @buf = buf
+      @buf.push Elem.new([lineno(), column()], __callee__, tok)
+    end
+
+    def on_heredoc_end(tok)
+      @buf.push Elem.new([lineno(), column()], __callee__, tok)
+      @buf = @stack.pop
+    end
+
+    def _push_token(tok)
+      @buf.push Elem.new([lineno(), column()], __callee__, tok)
+    end
+
+    (SCANNER_EVENTS.map {|event|:"on_#{event}"} - private_instance_methods(false)).each do |event|
+      alias_method event, :_push_token
     end
   end
 
@@ -90,9 +130,12 @@ class Ripper
 
   class TokenPattern   #:nodoc:
 
-    class Error < ::StandardError; end
-    class CompileError < Error; end
-    class MatchError < Error; end
+    class Error < ::StandardError # :nodoc:
+    end
+    class CompileError < Error # :nodoc:
+    end
+    class MatchError < Error # :nodoc:
+    end
 
     class << self
       alias compile new
@@ -155,7 +198,7 @@ class Ripper
       MAP[tok]  or raise CompileError, "unknown token: #{tok}"
     end
 
-    class MatchData
+    class MatchData # :nodoc:
       def initialize(tokens, match)
         @tokens = tokens
         @match = match
@@ -173,7 +216,7 @@ class Ripper
         @tokens[@match.begin(n)...@match.end(n)].map {|pos,type,str| str }
       end
     end
-  
+
   end
 
 end
