@@ -1,5 +1,6 @@
+# frozen_string_literal: false
 #
-# $Id: sexp.rb 36954 2012-09-12 23:04:41Z zzak $
+# $Id$
 #
 # Copyright (c) 2004,2005 Minero Aoki
 #
@@ -28,7 +29,9 @@ class Ripper
   #           [:bodystmt, [[:var_ref, [:@kw, "nil", [1, 9]]]], nil, nil, nil]]]]
   #
   def Ripper.sexp(src, filename = '-', lineno = 1)
-    SexpBuilderPP.new(src, filename, lineno).parse
+    builder = SexpBuilderPP.new(src, filename, lineno)
+    sexp = builder.parse
+    sexp unless builder.error?
   end
 
   # [EXPERIMENTAL]
@@ -52,48 +55,43 @@ class Ripper
   #             nil]]]]
   #
   def Ripper.sexp_raw(src, filename = '-', lineno = 1)
-    SexpBuilder.new(src, filename, lineno).parse
-  end
-
-  class SexpBuilderPP < ::Ripper   #:nodoc:
-    private
-
-    PARSER_EVENT_TABLE.each do |event, arity|
-      if /_new\z/ =~ event.to_s and arity == 0
-        module_eval(<<-End, __FILE__, __LINE__ + 1)
-          def on_#{event}
-            []
-          end
-        End
-      elsif /_add\z/ =~ event.to_s
-        module_eval(<<-End, __FILE__, __LINE__ + 1)
-          def on_#{event}(list, item)
-            list.push item
-            list
-          end
-        End
-      else
-        module_eval(<<-End, __FILE__, __LINE__ + 1)
-          def on_#{event}(*args)
-            [:#{event}, *args]
-          end
-        End
-      end
-    end
-
-    SCANNER_EVENTS.each do |event|
-      module_eval(<<-End, __FILE__, __LINE__ + 1)
-        def on_#{event}(tok)
-          [:@#{event}, tok, [lineno(), column()]]
-        end
-      End
-    end
+    builder = SexpBuilder.new(src, filename, lineno)
+    sexp = builder.parse
+    sexp unless builder.error?
   end
 
   class SexpBuilder < ::Ripper   #:nodoc:
     private
 
-    PARSER_EVENTS.each do |event|
+    def dedent_element(e, width)
+      if (n = dedent_string(e[1], width)) > 0
+        e[2][1] += n
+      end
+      e
+    end
+
+    def on_heredoc_dedent(val, width)
+      sub = proc do |cont|
+        cont.map! do |e|
+          if Array === e
+            case e[0]
+            when :@tstring_content
+              e = dedent_element(e, width)
+            when /_add\z/
+              e[1] = sub[e[1]]
+            end
+          elsif String === e
+            dedent_string(e, width)
+          end
+          e
+        end
+      end
+      sub[val]
+      val
+    end
+
+    events = private_instance_methods(false).grep(/\Aon_/) {$'.to_sym}
+    (PARSER_EVENTS - events).each do |event|
       module_eval(<<-End, __FILE__, __LINE__ + 1)
         def on_#{event}(*args)
           args.unshift :#{event}
@@ -111,4 +109,38 @@ class Ripper
     end
   end
 
+  class SexpBuilderPP < SexpBuilder #:nodoc:
+    private
+
+    def on_heredoc_dedent(val, width)
+      val.map! do |e|
+        next e if Symbol === e and /_content\z/ =~ e
+        if Array === e and e[0] == :@tstring_content
+          e = dedent_element(e, width)
+        elsif String === e
+          dedent_string(e, width)
+        end
+        e
+      end
+      val
+    end
+
+    def _dispatch_event_new
+      []
+    end
+
+    def _dispatch_event_push(list, item)
+      list.push item
+      list
+    end
+
+    PARSER_EVENT_TABLE.each do |event, arity|
+      if /_new\z/ =~ event and arity == 0
+        alias_method "on_#{event}", :_dispatch_event_new
+      elsif /_add\z/ =~ event
+        alias_method "on_#{event}", :_dispatch_event_push
+      end
+    end
+  end
+
 end
