@@ -1,5 +1,5 @@
 /*
-    $Id: strscan.c 44903 2014-02-10 11:45:14Z naruse $
+    $Id$
 
     Copyright (c) 1999-2006 Minero Aoki
 
@@ -64,6 +64,7 @@ struct strscanner
                             Function Prototypes
    ======================================================================= */
 
+static inline long minl _((const long n, const long x));
 static VALUE infect _((VALUE str, struct strscanner *p));
 static VALUE extract_range _((struct strscanner *p, long beg_i, long end_i));
 static VALUE extract_beg_len _((struct strscanner *p, long beg_i, long len));
@@ -140,12 +141,17 @@ str_new(struct strscanner *p, const char *ptr, long len)
     return str;
 }
 
+static inline long
+minl(const long x, const long y)
+{
+    return (x < y) ? x : y;
+}
+
 static VALUE
 extract_range(struct strscanner *p, long beg_i, long end_i)
 {
     if (beg_i > S_LEN(p)) return Qnil;
-    if (end_i > S_LEN(p))
-        end_i = S_LEN(p);
+    end_i = minl(end_i, S_LEN(p));
     return infect(str_new(p, S_PBEG(p) + beg_i, end_i - beg_i), p);
 }
 
@@ -153,8 +159,7 @@ static VALUE
 extract_beg_len(struct strscanner *p, long beg_i, long len)
 {
     if (beg_i > S_LEN(p)) return Qnil;
-    if (beg_i + len > S_LEN(p))
-        len = S_LEN(p) - beg_i;
+    len = minl(len, S_LEN(p) - beg_i);
     return infect(str_new(p, S_PBEG(p) + beg_i, len), p);
 }
 
@@ -181,30 +186,25 @@ static size_t
 strscan_memsize(const void *ptr)
 {
     const struct strscanner *p = ptr;
-    size_t size = 0;
-    if (p) {
-	size = sizeof(*p) - sizeof(p->regs) + onig_region_memsize(&p->regs);
-    }
-    return size;
+    return sizeof(*p) - sizeof(p->regs) + onig_region_memsize(&p->regs);
 }
 
 static const rb_data_type_t strscanner_type = {
     "StringScanner",
     {strscan_mark, strscan_free, strscan_memsize},
-    NULL, NULL, RUBY_TYPED_FREE_IMMEDIATELY
+    0, 0, RUBY_TYPED_FREE_IMMEDIATELY
 };
 
 static VALUE
 strscan_s_allocate(VALUE klass)
 {
     struct strscanner *p;
+    VALUE obj = TypedData_Make_Struct(klass, struct strscanner, &strscanner_type, p);
 
-    p = ALLOC(struct strscanner);
-    MEMZERO(p, struct strscanner, 1);
     CLEAR_MATCH_STATUS(p);
     onig_region_init(&(p->regs));
     p->str = Qnil;
-    return TypedData_Wrap_Struct(klass, &strscanner_type, p);
+    return obj;
 }
 
 /*
@@ -252,7 +252,9 @@ strscan_init_copy(VALUE vself, VALUE vorig)
 	self->str = orig->str;
 	self->prev = orig->prev;
 	self->curr = orig->curr;
-	onig_region_copy(&self->regs, &orig->regs);
+	if (rb_reg_region_copy(&self->regs, &orig->regs))
+	    rb_memerror();
+	RB_GC_GUARD(vorig);
     }
 
     return vself;
@@ -731,9 +733,7 @@ strscan_getch(VALUE self)
         return Qnil;
 
     len = rb_enc_mbclen(CURPTR(p), S_PEND(p), rb_enc_get(p->str));
-    if (p->curr + len > S_LEN(p)) {
-        len = S_LEN(p) - p->curr;
-    }
+    len = minl(len, S_RESTLEN(p));
     p->prev = p->curr;
     p->curr += len;
     MATCHED(p);
@@ -810,8 +810,7 @@ strscan_peek(VALUE self, VALUE vlen)
     if (EOS_P(p))
         return infect(str_new(p, "", 0), p);
 
-    if (p->curr + len > S_LEN(p))
-        len = S_LEN(p) - p->curr;
+    len = minl(len, S_RESTLEN(p));
     return extract_beg_len(p, p->curr, len);
 }
 
@@ -977,7 +976,7 @@ strscan_matched_size(VALUE self)
 }
 
 static int
-name_to_backref_number(struct re_registers *regs, VALUE regexp, const char* name, const char* name_end)
+name_to_backref_number(struct re_registers *regs, VALUE regexp, const char* name, const char* name_end, rb_encoding *enc)
 {
     int num;
 
@@ -987,9 +986,8 @@ name_to_backref_number(struct re_registers *regs, VALUE regexp, const char* name
 	return num;
     }
     else {
-	VALUE s = rb_str_new(name, (long )(name_end - name));
-	rb_raise(rb_eIndexError, "undefined group name reference: %s",
-				 StringValuePtr(s));
+	rb_enc_raise(enc, rb_eIndexError, "undefined group name reference: %.*s",
+					  rb_long2int(name_end - name), name);
     }
 
     UNREACHABLE;
@@ -1033,13 +1031,12 @@ strscan_aref(VALUE self, VALUE idx)
 
     switch (TYPE(idx)) {
         case T_SYMBOL:
-            name = rb_id2name(SYM2ID(idx));
-            goto name_to_backref;
-            break;
+            idx = rb_sym2str(idx);
+            /* fall through */
         case T_STRING:
-            name = StringValuePtr(idx);
-        name_to_backref:
-	    i = name_to_backref_number(&(p->regs), p->regex, name, name + strlen(name));
+            if (!p->regex) return Qnil;
+            RSTRING_GETMEM(idx, name, i);
+            i = name_to_backref_number(&(p->regs), p->regex, name, name + i, rb_enc_get(idx));
             break;
         default:
             i = NUM2LONG(idx);
@@ -1122,7 +1119,7 @@ strscan_rest_size(VALUE self)
     if (EOS_P(p)) {
         return INT2FIX(0);
     }
-    i = S_LEN(p) - p->curr;
+    i = S_RESTLEN(p);
     return INT2FIX(i);
 }
 
@@ -1138,7 +1135,6 @@ strscan_restsize(VALUE self)
 }
 
 #define INSPECT_LENGTH 5
-#define BUFSIZE 256
 
 /*
  * Returns a string that represents the StringScanner object, showing:
@@ -1209,7 +1205,7 @@ inspect2(struct strscanner *p)
     long len;
 
     if (EOS_P(p)) return rb_str_new2("");
-    len = S_LEN(p) - p->curr;
+    len = S_RESTLEN(p);
     if (len > INSPECT_LENGTH) {
 	str = rb_str_new(CURPTR(p), INSPECT_LENGTH);
 	rb_str_cat2(str, "...");
@@ -1328,7 +1324,7 @@ inspect2(struct strscanner *p)
  * There are aliases to several of the methods.
  */
 void
-Init_strscan()
+Init_strscan(void)
 {
     ID id_scanerr = rb_intern("ScanError");
     VALUE tmp;
@@ -1343,7 +1339,7 @@ Init_strscan()
     tmp = rb_str_new2(STRSCAN_VERSION);
     rb_obj_freeze(tmp);
     rb_const_set(StringScanner, rb_intern("Version"), tmp);
-    tmp = rb_str_new2("$Id: strscan.c 44903 2014-02-10 11:45:14Z naruse $");
+    tmp = rb_str_new2("$Id$");
     rb_obj_freeze(tmp);
     rb_const_set(StringScanner, rb_intern("Id"), tmp);
 
