@@ -1,3 +1,4 @@
+# frozen_string_literal: false
 #
 # = net/protocol.rb
 #
@@ -11,7 +12,7 @@
 # modify this program under the same terms as Ruby itself,
 # Ruby Distribute License or GNU General Public License.
 #
-# $Id: protocol.rb 38830 2013-01-15 07:23:48Z usa $
+# $Id$
 #++
 #
 # WARNING: This file is going to remove.
@@ -20,6 +21,7 @@
 
 require 'socket'
 require 'timeout'
+require 'io/wait'
 
 module Net # :nodoc:
 
@@ -32,6 +34,24 @@ module Net # :nodoc:
         end
       End
     end
+
+    def ssl_socket_connect(s, timeout)
+      if timeout
+        while true
+          raise Net::OpenTimeout if timeout <= 0
+          start = Process.clock_gettime Process::CLOCK_MONOTONIC
+          # to_io is required because SSLSocket doesn't have wait_readable yet
+          case s.connect_nonblock(exception: false)
+          when :wait_readable; s.to_io.wait_readable(timeout)
+          when :wait_writable; s.to_io.wait_writable(timeout)
+          else; break
+          end
+          timeout -= Process.clock_gettime(Process::CLOCK_MONOTONIC) - start
+        end
+      else
+        s.connect
+      end
+    end
   end
 
 
@@ -45,6 +65,18 @@ module Net # :nodoc:
   class ProtoRetriableError    < ProtocolError; end
   ProtocRetryError = ProtoRetriableError
 
+  ##
+  # OpenTimeout, a subclass of Timeout::Error, is raised if a connection cannot
+  # be created within the open_timeout.
+
+  class OpenTimeout            < Timeout::Error; end
+
+  ##
+  # ReadTimeout, a subclass of Timeout::Error, is raised if a chunk of the
+  # response cannot be read within the read_timeout.
+
+  class ReadTimeout            < Timeout::Error; end
+
 
   class BufferedIO   #:nodoc: internal use only
     def initialize(io)
@@ -137,23 +169,21 @@ module Net # :nodoc:
     BUFSIZE = 1024 * 16
 
     def rbuf_fill
-      begin
-        @rbuf << @io.read_nonblock(BUFSIZE)
-      rescue IO::WaitReadable
-        if IO.select([@io], nil, nil, @read_timeout)
-          retry
-        else
-          raise Timeout::Error
-        end
-      rescue IO::WaitWritable
+      case rv = @io.read_nonblock(BUFSIZE, exception: false)
+      when String
+        return @rbuf << rv
+      when :wait_readable
+        @io.to_io.wait_readable(@read_timeout) or raise Net::ReadTimeout
+        # continue looping
+      when :wait_writable
         # OpenSSL::Buffering#read_nonblock may fail with IO::WaitWritable.
         # http://www.openssl.org/support/faq.html#PROG10
-        if IO.select(nil, [@io], nil, @read_timeout)
-          retry
-        else
-          raise Timeout::Error
-        end
-      end
+        @io.to_io.wait_writable(@read_timeout) or raise Net::ReadTimeout
+        # continue looping
+      when nil
+        # callers do not care about backtrace, so avoid allocating for it
+        raise EOFError, 'end of file reached', []
+      end while true
     end
 
     def rbuf_consume(len)
@@ -255,7 +285,7 @@ module Net # :nodoc:
     def write_message_0(src)
       prev = @written_bytes
       each_crlf_line(src) do |line|
-        write0 line.sub(/\A\./, '..')
+        write0 dot_stuff(line)
       end
       @written_bytes - prev
     end
@@ -296,11 +326,15 @@ module Net # :nodoc:
 
     private
 
+    def dot_stuff(s)
+      s.sub(/\A\./, '..')
+    end
+
     def using_each_crlf_line
       @wbuf = ''
       yield
       if not @wbuf.empty?   # unterminated last line
-        write0 @wbuf.chomp + "\r\n"
+        write0 dot_stuff(@wbuf.chomp) + "\r\n"
       elsif @written_bytes == 0   # empty src
         write0 "\r\n"
       end
