@@ -1,3 +1,4 @@
+# frozen_string_literal: false
 # = net/smtp.rb
 #
 # Copyright (c) 1999-2007 Yukihiro Matsumoto.
@@ -14,7 +15,7 @@
 # NOTE: You can find Japanese version of this document at:
 # http://www.ruby-lang.org/ja/man/html/net_smtp.html
 #
-# $Id: smtp.rb 39018 2013-02-02 04:37:45Z zzak $
+# $Id$
 #
 # See Net::SMTP for documentation.
 #
@@ -76,8 +77,9 @@ module Net
   #
   # This library does NOT provide functions to compose internet mails.
   # You must create them by yourself. If you want better mail support,
-  # try RubyMail or TMail. You can get both libraries from RAA.
-  # (http://www.ruby-lang.org/en/raa.html)
+  # try RubyMail or TMail or search for alternatives in
+  # {RubyGems.org}[https://rubygems.org/] or {The Ruby
+  # Toolbox}[https://www.ruby-toolbox.com/].
   #
   # FYI: the official documentation on internet mail is: [RFC2822] (http://www.ietf.org/rfc/rfc2822.txt).
   #
@@ -168,9 +170,9 @@ module Net
   #     Net::SMTP.start('your.smtp.server', 25, 'mail.from.domain',
   #                     'Your Account', 'Your Password', :cram_md5)
   #
-  class SMTP
+  class SMTP < Protocol
 
-    Revision = %q$Revision: 39018 $.split[1]
+    Revision = %q$Revision$.split[1]
 
     # The default SMTP port number, 25.
     def SMTP.default_port
@@ -215,7 +217,7 @@ module Net
       @started = false
       @open_timeout = 30
       @read_timeout = 60
-      @error_occured = false
+      @error_occurred = false
       @debug_output = nil
       @tls = false
       @starttls = false
@@ -362,12 +364,12 @@ module Net
 
     # Seconds to wait while attempting to open a connection.
     # If the connection cannot be opened within this time, a
-    # Net::OpenTimeout is raised.
+    # Net::OpenTimeout is raised. The default value is 30 seconds.
     attr_accessor :open_timeout
 
     # Seconds to wait while reading one block (by one read(2) call).
     # If the read(2) call does not complete within this time, a
-    # Net::ReadTimeout is raised.
+    # Net::ReadTimeout is raised. The default value is 60 seconds.
     attr_reader :read_timeout
 
     # Set the number of seconds to wait until timing-out a read(2)
@@ -582,7 +584,7 @@ module Net
       s = ssl_socket(s, @ssl_context)
       logging "TLS connection started"
       s.sync_close = true
-      s.connect
+      ssl_socket_connect(s, @open_timeout)
       if @ssl_context.verify_mode != OpenSSL::SSL::VERIFY_NONE
         s.post_connection_check(@address)
       end
@@ -605,17 +607,17 @@ module Net
     rescue SMTPError
       if @esmtp
         @esmtp = false
-        @error_occured = false
+        @error_occurred = false
         retry
       end
       raise
     end
 
     def do_finish
-      quit if @socket and not @socket.closed? and not @error_occured
+      quit if @socket and not @socket.closed? and not @error_occurred
     ensure
       @started = false
-      @error_occured = false
+      @error_occurred = false
       @socket.close if @socket and not @socket.closed?
       @socket = nil
     end
@@ -815,6 +817,12 @@ module Net
 
     public
 
+    # Aborts the current mail transaction
+
+    def rset
+      getok('RSET')
+    end
+
     def starttls
       getok('STARTTLS')
     end
@@ -894,10 +902,17 @@ module Net
       end
       res = critical {
         check_continue get_response('DATA')
-        if msgstr
-          @socket.write_message msgstr
-        else
-          @socket.write_message_by_block(&block)
+        socket_sync_bak = @socket.io.sync
+        begin
+          @socket.io.sync = false
+          if msgstr
+            @socket.write_message msgstr
+          else
+            @socket.write_message_by_block(&block)
+          end
+        ensure
+          @socket.io.flush
+          @socket.io.sync = socket_sync_bak
         end
         recv_response()
       }
@@ -911,7 +926,15 @@ module Net
 
     private
 
+    def validate_line(line)
+      # A bare CR or LF is not allowed in RFC5321.
+      if /[\r\n]/ =~ line
+        raise ArgumentError, "A line must not contain CR or LF"
+      end
+    end
+
     def getok(reqline)
+      validate_line reqline
       res = critical {
         @socket.writeline reqline
         recv_response()
@@ -921,6 +944,7 @@ module Net
     end
 
     def get_response(reqline)
+      validate_line reqline
       @socket.writeline reqline
       recv_response()
     end
@@ -936,11 +960,11 @@ module Net
     end
 
     def critical
-      return '200 dummy reply code' if @error_occured
+      return Response.parse('200 dummy reply code') if @error_occurred
       begin
         return yield()
       rescue Exception
-        @error_occured = true
+        @error_occurred = true
         raise
       end
     end
@@ -1035,7 +1059,7 @@ module Net
         h
       end
 
-      # Determines whether there was an error and raies the appropriate error
+      # Determines whether there was an error and raises the appropriate error
       # based on the reply code of the response
       def exception_class
         case @status
