@@ -9,8 +9,6 @@
 
 **********************************************************************/
 
-#include "ruby/ruby.h"
-#include "ruby/encoding.h"
 #include "internal.h"
 #include <sys/types.h>
 #include <ctype.h>
@@ -25,11 +23,11 @@
  * This behavior is consistent with the document of pack/unpack.
  */
 #ifdef HAVE_TRUE_LONG_LONG
-static const char natstr[] = "sSiIlLqQ";
+static const char natstr[] = "sSiIlLqQjJ";
 #else
-static const char natstr[] = "sSiIlL";
+static const char natstr[] = "sSiIlLjJ";
 #endif
-static const char endstr[] = "sSiIlLqQ";
+static const char endstr[] = "sSiIlLqQjJ";
 
 #ifdef HAVE_TRUE_LONG_LONG
 /* It is intentional to use long long instead of LONG_LONG. */
@@ -70,99 +68,18 @@ static const char endstr[] = "sSiIlLqQ";
 # define NATINT_LEN(type,len) ((int)sizeof(type))
 #endif
 
-#if SIZEOF_LONG == 8
-# define INT64toNUM(x) LONG2NUM(x)
-# define UINT64toNUM(x) ULONG2NUM(x)
-#elif defined(HAVE_LONG_LONG) && SIZEOF_LONG_LONG == 8
-# define INT64toNUM(x) LL2NUM(x)
-# define UINT64toNUM(x) ULL2NUM(x)
-#endif
-
-#define define_swapx(x, xtype)		\
-static xtype				\
-TOKEN_PASTE(swap,x)(xtype z)		\
-{					\
-    xtype r;				\
-    xtype *zp;				\
-    unsigned char *s, *t;		\
-    int i;				\
-					\
-    zp = xmalloc(sizeof(xtype));	\
-    *zp = z;				\
-    s = (unsigned char*)zp;		\
-    t = xmalloc(sizeof(xtype));		\
-    for (i=0; i<sizeof(xtype); i++) {	\
-	t[sizeof(xtype)-i-1] = s[i];	\
-    }					\
-    r = *(xtype *)t;			\
-    xfree(t);				\
-    xfree(zp);				\
-    return r;				\
-}
-
-#if SIZEOF_SHORT == 2
-# define swaps(x)	swap16(x)
-#elif SIZEOF_SHORT == 4
-# define swaps(x)	swap32(x)
-#else
-  define_swapx(s,short)
-#endif
-
-#if SIZEOF_INT == 2
-# define swapi(x)	swap16(x)
-#elif SIZEOF_INT == 4
-# define swapi(x)	swap32(x)
-#else
-  define_swapx(i,int)
-#endif
-
-#if SIZEOF_LONG == 4
-# define swapl(x)	swap32(x)
-#elif SIZEOF_LONG == 8
-# define swapl(x)        swap64(x)
-#else
-  define_swapx(l,long)
-#endif
-
-#ifdef HAVE_LONG_LONG
-# if SIZEOF_LONG_LONG == 8
-#  define swapll(x)        swap64(x)
-# else
-   define_swapx(ll,LONG_LONG)
-# endif
-#endif
-
-#if SIZEOF_FLOAT == 4 && defined(HAVE_INT32_T)
-#   define swapf(x)	swap32(x)
-#   define FLOAT_SWAPPER	uint32_t
-#else
-    define_swapx(f,float)
-#endif
-
-#if SIZEOF_DOUBLE == 8 && defined(HAVE_INT64_T)
-#   define swapd(x)	swap64(x)
-#   define DOUBLE_SWAPPER	uint64_t
-#elif SIZEOF_DOUBLE == 8 && defined(HAVE_INT32_T)
-    static double
-    swapd(const double d)
-    {
-	double dtmp = d;
-	uint32_t utmp[2];
-	uint32_t utmp0;
-
-	utmp[0] = 0; utmp[1] = 0;
-	memcpy(utmp,&dtmp,sizeof(double));
-	utmp0 = utmp[0];
-	utmp[0] = swap32(utmp[1]);
-	utmp[1] = swap32(utmp0);
-	memcpy(&dtmp,utmp,sizeof(double));
-	return dtmp;
-    }
-#else
-    define_swapx(d, double)
-#endif
-
-#undef define_swapx
+typedef union {
+    float f;
+    uint32_t u;
+    char buf[4];
+} FLOAT_SWAPPER;
+typedef union {
+    double d;
+    uint64_t u;
+    char buf[8];
+} DOUBLE_SWAPPER;
+#define swapf(x) swap32(x)
+#define swapd(x) swap64(x)
 
 #define rb_ntohf(x) (BIGENDIAN_P()?(x):swapf(x))
 #define rb_ntohd(x) (BIGENDIAN_P()?(x):swapd(x))
@@ -173,57 +90,17 @@ TOKEN_PASTE(swap,x)(xtype z)		\
 #define rb_vtohf(x) (BIGENDIAN_P()?swapf(x):(x))
 #define rb_vtohd(x) (BIGENDIAN_P()?swapd(x):(x))
 
-#ifdef FLOAT_SWAPPER
-# define FLOAT_CONVWITH(y)	FLOAT_SWAPPER y;
-# define HTONF(x,y)	(memcpy(&(y),&(x),sizeof(float)),	\
-			 (y) = rb_htonf((FLOAT_SWAPPER)(y)),	\
-			 memcpy(&(x),&(y),sizeof(float)),	\
-			 (x))
-# define HTOVF(x,y)	(memcpy(&(y),&(x),sizeof(float)),	\
-			 (y) = rb_htovf((FLOAT_SWAPPER)(y)),	\
-			 memcpy(&(x),&(y),sizeof(float)),	\
-			 (x))
-# define NTOHF(x,y)	(memcpy(&(y),&(x),sizeof(float)),	\
-			 (y) = rb_ntohf((FLOAT_SWAPPER)(y)),	\
-			 memcpy(&(x),&(y),sizeof(float)),	\
-			 (x))
-# define VTOHF(x,y)	(memcpy(&(y),&(x),sizeof(float)),	\
-			 (y) = rb_vtohf((FLOAT_SWAPPER)(y)),	\
-			 memcpy(&(x),&(y),sizeof(float)),	\
-			 (x))
-#else
-# define FLOAT_CONVWITH(y)
-# define HTONF(x,y)	rb_htonf(x)
-# define HTOVF(x,y)	rb_htovf(x)
-# define NTOHF(x,y)	rb_ntohf(x)
-# define VTOHF(x,y)	rb_vtohf(x)
-#endif
+#define FLOAT_CONVWITH(x)	FLOAT_SWAPPER x;
+#define HTONF(x)	((x).u = rb_htonf((x).u))
+#define HTOVF(x)	((x).u = rb_htovf((x).u))
+#define NTOHF(x)	((x).u = rb_ntohf((x).u))
+#define VTOHF(x)	((x).u = rb_vtohf((x).u))
 
-#ifdef DOUBLE_SWAPPER
-# define DOUBLE_CONVWITH(y)	DOUBLE_SWAPPER y;
-# define HTOND(x,y)	(memcpy(&(y),&(x),sizeof(double)),	\
-			 (y) = rb_htond((DOUBLE_SWAPPER)(y)),	\
-			 memcpy(&(x),&(y),sizeof(double)),	\
-			 (x))
-# define HTOVD(x,y)	(memcpy(&(y),&(x),sizeof(double)),	\
-			 (y) = rb_htovd((DOUBLE_SWAPPER)(y)),	\
-			 memcpy(&(x),&(y),sizeof(double)),	\
-			 (x))
-# define NTOHD(x,y)	(memcpy(&(y),&(x),sizeof(double)),	\
-			 (y) = rb_ntohd((DOUBLE_SWAPPER)(y)),	\
-			 memcpy(&(x),&(y),sizeof(double)),	\
-			 (x))
-# define VTOHD(x,y)	(memcpy(&(y),&(x),sizeof(double)),	\
-			 (y) = rb_vtohd((DOUBLE_SWAPPER)(y)),	\
-			 memcpy(&(x),&(y),sizeof(double)),	\
-			 (x))
-#else
-# define DOUBLE_CONVWITH(y)
-# define HTOND(x,y)	rb_htond(x)
-# define HTOVD(x,y)	rb_htovd(x)
-# define NTOHD(x,y)	rb_ntohd(x)
-# define VTOHD(x,y)	rb_vtohd(x)
-#endif
+#define DOUBLE_CONVWITH(x)	DOUBLE_SWAPPER x;
+#define HTOND(x)	((x).u = rb_htond((x).u))
+#define HTOVD(x)	((x).u = rb_htovd((x).u))
+#define NTOHD(x)	((x).u = rb_ntohd((x).u))
+#define VTOHD(x)	((x).u = rb_vtohd((x).u))
 
 #define MAX_INTEGER_PACK_SIZE 8
 
@@ -234,6 +111,45 @@ static void qpencode(VALUE,VALUE,long);
 
 static unsigned long utf8_to_uv(const char*,long*);
 
+static ID id_associated;
+
+static void
+str_associate(VALUE str, VALUE add)
+{
+    VALUE assoc;
+
+    assoc = rb_attr_get(str, id_associated);
+    if (RB_TYPE_P(assoc, T_ARRAY)) {
+	/* already associated */
+	rb_ary_concat(assoc, add);
+    }
+    else {
+	rb_ivar_set(str, id_associated, add);
+    }
+}
+
+static VALUE
+str_associated(VALUE str)
+{
+    VALUE assoc = rb_attr_get(str, id_associated);
+    if (NIL_P(assoc)) assoc = Qfalse;
+    return assoc;
+}
+
+void
+rb_str_associate(VALUE str, VALUE add)
+{
+    rb_warn("rb_str_associate() is only for internal use and deprecated; do not use");
+    str_associate(str, add);
+}
+
+VALUE
+rb_str_associated(VALUE str)
+{
+    rb_warn("rb_str_associated() is only for internal use and deprecated; do not use");
+    return str_associated(str);
+}
+
 /*
  *  call-seq:
  *     arr.pack ( aTemplateString ) -> aBinaryString
@@ -267,11 +183,15 @@ static unsigned long utf8_to_uv(const char*,long*);
  *      S         | Integer | 16-bit unsigned, native endian (uint16_t)
  *      L         | Integer | 32-bit unsigned, native endian (uint32_t)
  *      Q         | Integer | 64-bit unsigned, native endian (uint64_t)
+ *      J         | Integer | pointer width unsigned, native endian (uintptr_t)
+ *                |         | (J is available since Ruby 2.3.)
  *                |         |
  *      c         | Integer | 8-bit signed (signed char)
  *      s         | Integer | 16-bit signed, native endian (int16_t)
  *      l         | Integer | 32-bit signed, native endian (int32_t)
  *      q         | Integer | 64-bit signed, native endian (int64_t)
+ *      j         | Integer | pointer width signed, native endian (intptr_t)
+ *                |         | (j is available since Ruby 2.3.)
  *                |         |
  *      S_, S!    | Integer | unsigned short, native endian
  *      I, I_, I! | Integer | unsigned int, native endian
@@ -279,6 +199,8 @@ static unsigned long utf8_to_uv(const char*,long*);
  *      Q_, Q!    | Integer | unsigned long long, native endian (ArgumentError
  *                |         | if the platform has no long long type.)
  *                |         | (Q_ and Q! is available since Ruby 2.1.)
+ *      J!        | Integer | uintptr_t, native endian (same with J)
+ *                |         | (J! is available since Ruby 2.3.)
  *                |         |
  *      s_, s!    | Integer | signed short, native endian
  *      i, i_, i! | Integer | signed int, native endian
@@ -286,20 +208,26 @@ static unsigned long utf8_to_uv(const char*,long*);
  *      q_, q!    | Integer | signed long long, native endian (ArgumentError
  *                |         | if the platform has no long long type.)
  *                |         | (q_ and q! is available since Ruby 2.1.)
+ *      j!        | Integer | intptr_t, native endian (same with j)
+ *                |         | (j! is available since Ruby 2.3.)
  *                |         |
  *      S> L> Q>  | Integer | same as the directives without ">" except
- *      s> l> q>  |         | big endian
- *      S!> I!>   |         | (available since Ruby 1.9.3)
- *      L!> Q!>   |         | "S>" is same as "n"
- *      s!> i!>   |         | "L>" is same as "N"
- *      l!> q!>   |         |
+ *      J> s> l>  |         | big endian
+ *      q> j>     |         | (available since Ruby 1.9.3)
+ *      S!> I!>   |         | "S>" is same as "n"
+ *      L!> Q!>   |         | "L>" is same as "N"
+ *      J!> s!>   |         |
+ *      i!> l!>   |         |
+ *      q!> j!>   |         |
  *                |         |
  *      S< L< Q<  | Integer | same as the directives without "<" except
- *      s< l< q<  |         | little endian
- *      S!< I!<   |         | (available since Ruby 1.9.3)
- *      L!< Q!<   |         | "S<" is same as "v"
- *      s!< i!<   |         | "L<" is same as "V"
- *      l!< q!<   |         |
+ *      J< s< l<  |         | little endian
+ *      q< j<     |         | (available since Ruby 1.9.3)
+ *      S!< I!<   |         | "S<" is same as "v"
+ *      L!< Q!<   |         | "L<" is same as "V"
+ *      J!< s!<   |         |
+ *      i!< l!<   |         |
+ *      q!< j!<   |         |
  *                |         |
  *      n         | Integer | 16-bit unsigned, network (big-endian) byte order
  *      N         | Integer | 32-bit unsigned, network (big-endian) byte order
@@ -352,7 +280,7 @@ pack_pack(VALUE ary, VALUE fmt)
     const char *p, *pend;
     VALUE res, from, associates = 0;
     char type;
-    long items, len, idx, plen;
+    long len, idx, plen;
     const char *ptr;
     int enc_info = 1;		/* 0 - BINARY, 1 - US-ASCII, 2 - UTF-8 */
 #ifdef NATINT_PACK
@@ -365,12 +293,12 @@ pack_pack(VALUE ary, VALUE fmt)
     pend = p + RSTRING_LEN(fmt);
     res = rb_str_buf_new(0);
 
-    items = RARRAY_LEN(ary);
     idx = 0;
 
 #define TOO_FEW (rb_raise(rb_eArgError, toofew), 0)
-#define THISFROM (items > 0 ? RARRAY_AREF(ary, idx) : TOO_FEW)
-#define NEXTFROM (items-- > 0 ? RARRAY_AREF(ary, idx++) : TOO_FEW)
+#define MORE_ITEM (idx < RARRAY_LEN(ary))
+#define THISFROM (MORE_ITEM ? RARRAY_AREF(ary, idx) : TOO_FEW)
+#define NEXTFROM (MORE_ITEM ? RARRAY_AREF(ary, idx++) : TOO_FEW)
 
     while (p < pend) {
 	int explicit_endian = 0;
@@ -422,7 +350,7 @@ pack_pack(VALUE ary, VALUE fmt)
 	if (*p == '*') {	/* set data length */
 	    len = strchr("@Xxu", type) ? 0
                 : strchr("PMm", type) ? 1
-                : items;
+                : RARRAY_LEN(ary) - idx;
 	    p++;
 	}
 	else if (ISDIGIT(*p)) {
@@ -661,6 +589,16 @@ pack_pack(VALUE ary, VALUE fmt)
             bigendian_p = BIGENDIAN_P();
             goto pack_integer;
 
+	  case 'j':		/* j for intptr_t */
+	    integer_size = sizeof(intptr_t);
+	    bigendian_p = BIGENDIAN_P();
+	    goto pack_integer;
+
+	  case 'J':		/* J for uintptr_t */
+	    integer_size = sizeof(uintptr_t);
+	    bigendian_p = BIGENDIAN_P();
+	    goto pack_integer;
+
 	  case 'n':		/* 16 bit (2 bytes) integer (network byte-order)  */
             integer_size = 2;
             bigendian_p = 1;
@@ -711,25 +649,22 @@ pack_pack(VALUE ary, VALUE fmt)
 
 	  case 'e':		/* single precision float in VAX byte-order */
 	    while (len-- > 0) {
-		float f;
-		FLOAT_CONVWITH(ftmp);
+		FLOAT_CONVWITH(tmp);
 
 		from = NEXTFROM;
-		f = (float)RFLOAT_VALUE(rb_to_float(from));
-		f = HTOVF(f,ftmp);
-		rb_str_buf_cat(res, (char*)&f, sizeof(float));
+		tmp.f = (float)RFLOAT_VALUE(rb_to_float(from));
+		HTOVF(tmp);
+		rb_str_buf_cat(res, tmp.buf, sizeof(float));
 	    }
 	    break;
 
 	  case 'E':		/* double precision float in VAX byte-order */
 	    while (len-- > 0) {
-		double d;
-		DOUBLE_CONVWITH(dtmp);
-
+		DOUBLE_CONVWITH(tmp);
 		from = NEXTFROM;
-		d = RFLOAT_VALUE(rb_to_float(from));
-		d = HTOVD(d,dtmp);
-		rb_str_buf_cat(res, (char*)&d, sizeof(double));
+		tmp.d = RFLOAT_VALUE(rb_to_float(from));
+		HTOVD(tmp);
+		rb_str_buf_cat(res, tmp.buf, sizeof(double));
 	    }
 	    break;
 
@@ -746,25 +681,22 @@ pack_pack(VALUE ary, VALUE fmt)
 
 	  case 'g':		/* single precision float in network byte-order */
 	    while (len-- > 0) {
-		float f;
-		FLOAT_CONVWITH(ftmp);
-
+		FLOAT_CONVWITH(tmp);
 		from = NEXTFROM;
-		f = (float)RFLOAT_VALUE(rb_to_float(from));
-		f = HTONF(f,ftmp);
-		rb_str_buf_cat(res, (char*)&f, sizeof(float));
+		tmp.f = (float)RFLOAT_VALUE(rb_to_float(from));
+		HTONF(tmp);
+		rb_str_buf_cat(res, tmp.buf, sizeof(float));
 	    }
 	    break;
 
 	  case 'G':		/* double precision float in network byte-order */
 	    while (len-- > 0) {
-		double d;
-		DOUBLE_CONVWITH(dtmp);
+		DOUBLE_CONVWITH(tmp);
 
 		from = NEXTFROM;
-		d = RFLOAT_VALUE(rb_to_float(from));
-		d = HTOND(d,dtmp);
-		rb_str_buf_cat(res, (char*)&d, sizeof(double));
+		tmp.d = RFLOAT_VALUE(rb_to_float(from));
+		HTOND(tmp);
+		rb_str_buf_cat(res, tmp.buf, sizeof(double));
 	    }
 	    break;
 
@@ -871,12 +803,12 @@ pack_pack(VALUE ary, VALUE fmt)
 		}
 		else {
 		    t = StringValuePtr(from);
+		    rb_obj_taint(from);
 		}
 		if (!associates) {
 		    associates = rb_ary_new();
 		}
 		rb_ary_push(associates, from);
-		rb_obj_taint(from);
 		rb_str_buf_cat(res, (char*)&t, sizeof(char*));
 	    }
 	    break;
@@ -913,15 +845,24 @@ pack_pack(VALUE ary, VALUE fmt)
 	    }
 	    break;
 
-	  default:
-	    rb_warning("unknown pack directive '%c' in '%s'",
-		type, RSTRING_PTR(fmt));
+	  default: {
+	    char unknown[5];
+	    if (ISPRINT(type)) {
+		unknown[0] = type;
+		unknown[1] = '\0';
+	    }
+	    else {
+		snprintf(unknown, sizeof(unknown), "\\x%.2x", type & 0xff);
+	    }
+	    rb_warning("unknown pack directive '%s' in '% "PRIsVALUE"'",
+		       unknown, fmt);
 	    break;
+	  }
 	}
     }
 
     if (associates) {
-	rb_str_associate(res, associates);
+	str_associate(res, associates);
     }
     OBJ_INFECT(res, fmt);
     switch (enc_info) {
@@ -944,12 +885,14 @@ static const char b64_table[] =
 "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
 
 static void
-encodes(VALUE str, const char *s, long len, int type, int tail_lf)
+encodes(VALUE str, const char *s0, long len, int type, int tail_lf)
 {
-    char buff[4096];
+    enum {buff_size = 4096, encoded_unit = 4, input_unit = 3};
+    char buff[buff_size + 1];	/* +1 for tail_lf */
     long i = 0;
-    const char *trans = type == 'u' ? uu_table : b64_table;
+    const char *const trans = type == 'u' ? uu_table : b64_table;
     char padding;
+    const unsigned char *s = (const unsigned char *)s0;
 
     if (type == 'u') {
 	buff[i++] = (char)len + ' ';
@@ -958,16 +901,16 @@ encodes(VALUE str, const char *s, long len, int type, int tail_lf)
     else {
 	padding = '=';
     }
-    while (len >= 3) {
-        while (len >= 3 && sizeof(buff)-i >= 4) {
+    while (len >= input_unit) {
+        while (len >= input_unit && buff_size-i >= encoded_unit) {
             buff[i++] = trans[077 & (*s >> 2)];
             buff[i++] = trans[077 & (((*s << 4) & 060) | ((s[1] >> 4) & 017))];
             buff[i++] = trans[077 & (((s[1] << 2) & 074) | ((s[2] >> 6) & 03))];
             buff[i++] = trans[077 & s[2]];
-            s += 3;
-            len -= 3;
+            s += input_unit;
+            len -= input_unit;
         }
-        if (sizeof(buff)-i < 4) {
+        if (buff_size-i < encoded_unit) {
             rb_str_buf_cat(str, buff, i);
             i = 0;
         }
@@ -987,6 +930,7 @@ encodes(VALUE str, const char *s, long len, int type, int tail_lf)
     }
     if (tail_lf) buff[i++] = '\n';
     rb_str_buf_cat(str, buff, i);
+    if ((size_t)i > sizeof(buff)) rb_bug("encodes() buffer overrun");
 }
 
 static const char hex_table[] = "0123456789ABCDEF";
@@ -1069,6 +1013,15 @@ hex2num(char c)
 	rb_ary_store(ary, RARRAY_LEN(ary)+tmp_len-1, Qnil); \
 } while (0)
 
+/* Workaround for Oracle Solaris Studio 12.4 C compiler optimization bug
+ * with "-xO4" optimization option.
+ */
+#if defined(__SUNPRO_C) && __SUNPRO_C == 0x5130
+# define AVOID_CC_BUG volatile
+#else
+# define AVOID_CC_BUG
+#endif
+
 static VALUE
 infected_str_new(const char *ptr, long len, VALUE str)
 {
@@ -1195,12 +1148,13 @@ infected_str_new(const char *ptr, long len, VALUE str)
 static VALUE
 pack_unpack(VALUE str, VALUE fmt)
 {
-    static const char hexdigits[] = "0123456789abcdef";
+#define hexdigits ruby_hexdigits
     char *s, *send;
     char *p, *pend;
     VALUE ary;
     char type;
-    long len, tmp_len;
+    long len;
+    AVOID_CC_BUG long tmp_len;
     int star;
 #ifdef NATINT_PACK
     int natint;			/* native integer */
@@ -1337,13 +1291,14 @@ pack_unpack(VALUE str, VALUE fmt)
 		if (p[-1] == '*' || len > (send - s) * 8)
 		    len = (send - s) * 8;
 		bits = 0;
-		UNPACK_PUSH(bitstr = rb_usascii_str_new(0, len));
+		bitstr = rb_usascii_str_new(0, len);
 		t = RSTRING_PTR(bitstr);
 		for (i=0; i<len; i++) {
 		    if (i & 7) bits >>= 1;
-		    else bits = *s++;
+		    else bits = (unsigned char)*s++;
 		    *t++ = (bits & 1) ? '1' : '0';
 		}
+		UNPACK_PUSH(bitstr);
 	    }
 	    break;
 
@@ -1357,13 +1312,14 @@ pack_unpack(VALUE str, VALUE fmt)
 		if (p[-1] == '*' || len > (send - s) * 8)
 		    len = (send - s) * 8;
 		bits = 0;
-		UNPACK_PUSH(bitstr = rb_usascii_str_new(0, len));
+		bitstr = rb_usascii_str_new(0, len);
 		t = RSTRING_PTR(bitstr);
 		for (i=0; i<len; i++) {
 		    if (i & 7) bits <<= 1;
-		    else bits = *s++;
+		    else bits = (unsigned char)*s++;
 		    *t++ = (bits & 128) ? '1' : '0';
 		}
+		UNPACK_PUSH(bitstr);
 	    }
 	    break;
 
@@ -1377,15 +1333,16 @@ pack_unpack(VALUE str, VALUE fmt)
 		if (p[-1] == '*' || len > (send - s) * 2)
 		    len = (send - s) * 2;
 		bits = 0;
-		UNPACK_PUSH(bitstr = rb_usascii_str_new(0, len));
+		bitstr = rb_usascii_str_new(0, len);
 		t = RSTRING_PTR(bitstr);
 		for (i=0; i<len; i++) {
 		    if (i & 1)
 			bits >>= 4;
 		    else
-			bits = *s++;
+			bits = (unsigned char)*s++;
 		    *t++ = hexdigits[bits & 15];
 		}
+		UNPACK_PUSH(bitstr);
 	    }
 	    break;
 
@@ -1399,15 +1356,16 @@ pack_unpack(VALUE str, VALUE fmt)
 		if (p[-1] == '*' || len > (send - s) * 2)
 		    len = (send - s) * 2;
 		bits = 0;
-		UNPACK_PUSH(bitstr = rb_usascii_str_new(0, len));
+		bitstr = rb_usascii_str_new(0, len);
 		t = RSTRING_PTR(bitstr);
 		for (i=0; i<len; i++) {
 		    if (i & 1)
 			bits <<= 4;
 		    else
-			bits = *s++;
+			bits = (unsigned char)*s++;
 		    *t++ = hexdigits[(bits >> 4) & 15];
 		}
+		UNPACK_PUSH(bitstr);
 	    }
 	    break;
 
@@ -1471,6 +1429,18 @@ pack_unpack(VALUE str, VALUE fmt)
 	    bigendian_p = BIGENDIAN_P();
 	    goto unpack_integer;
 
+	  case 'j':
+	    signed_p = 1;
+	    integer_size = sizeof(intptr_t);
+	    bigendian_p = BIGENDIAN_P();
+	    goto unpack_integer;
+
+	  case 'J':
+	    signed_p = 0;
+	    integer_size = sizeof(uintptr_t);
+	    bigendian_p = BIGENDIAN_P();
+	    goto unpack_integer;
+
 	  case 'n':
 	    signed_p = 0;
 	    integer_size = 2;
@@ -1527,13 +1497,11 @@ pack_unpack(VALUE str, VALUE fmt)
 	  case 'e':
 	    PACK_LENGTH_ADJUST_SIZE(sizeof(float));
 	    while (len-- > 0) {
-	        float tmp;
-		FLOAT_CONVWITH(ftmp);
-
-		memcpy(&tmp, s, sizeof(float));
+		FLOAT_CONVWITH(tmp);
+		memcpy(tmp.buf, s, sizeof(float));
 		s += sizeof(float);
-		tmp = VTOHF(tmp,ftmp);
-		UNPACK_PUSH(DBL2NUM((double)tmp));
+		VTOHF(tmp);
+		UNPACK_PUSH(DBL2NUM(tmp.f));
 	    }
 	    PACK_ITEM_ADJUST();
 	    break;
@@ -1541,13 +1509,11 @@ pack_unpack(VALUE str, VALUE fmt)
 	  case 'E':
 	    PACK_LENGTH_ADJUST_SIZE(sizeof(double));
 	    while (len-- > 0) {
-		double tmp;
-		DOUBLE_CONVWITH(dtmp);
-
-		memcpy(&tmp, s, sizeof(double));
+		DOUBLE_CONVWITH(tmp);
+		memcpy(tmp.buf, s, sizeof(double));
 		s += sizeof(double);
-		tmp = VTOHD(tmp,dtmp);
-		UNPACK_PUSH(DBL2NUM(tmp));
+		VTOHD(tmp);
+		UNPACK_PUSH(DBL2NUM(tmp.d));
 	    }
 	    PACK_ITEM_ADJUST();
 	    break;
@@ -1567,13 +1533,11 @@ pack_unpack(VALUE str, VALUE fmt)
 	  case 'g':
 	    PACK_LENGTH_ADJUST_SIZE(sizeof(float));
 	    while (len-- > 0) {
-	        float tmp;
-		FLOAT_CONVWITH(ftmp);
-
-		memcpy(&tmp, s, sizeof(float));
+		FLOAT_CONVWITH(tmp);
+		memcpy(tmp.buf, s, sizeof(float));
 		s += sizeof(float);
-		tmp = NTOHF(tmp,ftmp);
-		UNPACK_PUSH(DBL2NUM((double)tmp));
+		NTOHF(tmp);
+		UNPACK_PUSH(DBL2NUM(tmp.f));
 	    }
 	    PACK_ITEM_ADJUST();
 	    break;
@@ -1581,13 +1545,11 @@ pack_unpack(VALUE str, VALUE fmt)
 	  case 'G':
 	    PACK_LENGTH_ADJUST_SIZE(sizeof(double));
 	    while (len-- > 0) {
-		double tmp;
-		DOUBLE_CONVWITH(dtmp);
-
-		memcpy(&tmp, s, sizeof(double));
+		DOUBLE_CONVWITH(tmp);
+		memcpy(tmp.buf, s, sizeof(double));
 		s += sizeof(double);
-		tmp = NTOHD(tmp,dtmp);
-		UNPACK_PUSH(DBL2NUM(tmp));
+		NTOHD(tmp);
+		UNPACK_PUSH(DBL2NUM(tmp.d));
 	    }
 	    PACK_ITEM_ADJUST();
 	    break;
@@ -1610,12 +1572,12 @@ pack_unpack(VALUE str, VALUE fmt)
 		char *ptr = RSTRING_PTR(buf);
 		long total = 0;
 
-		while (s < send && *s > ' ' && *s < 'a') {
+		while (s < send && (unsigned char)*s > ' ' && (unsigned char)*s < 'a') {
 		    long a,b,c,d;
-		    char hunk[4];
+		    char hunk[3];
+
+		    len = ((unsigned char)*s++ - ' ') & 077;
 
-		    hunk[3] = '\0';
-		    len = (*s++ - ' ') & 077;
 		    total += len;
 		    if (total > RSTRING_LEN(buf)) {
 			len -= total - RSTRING_LEN(buf);
@@ -1625,20 +1587,20 @@ pack_unpack(VALUE str, VALUE fmt)
 		    while (len > 0) {
 			long mlen = len > 3 ? 3 : len;
 
-			if (s < send && *s >= ' ')
-			    a = (*s++ - ' ') & 077;
+			if (s < send && (unsigned char)*s >= ' ' && (unsigned char)*s < 'a')
+			    a = ((unsigned char)*s++ - ' ') & 077;
 			else
 			    a = 0;
-			if (s < send && *s >= ' ')
-			    b = (*s++ - ' ') & 077;
+			if (s < send && (unsigned char)*s >= ' ' && (unsigned char)*s < 'a')
+			    b = ((unsigned char)*s++ - ' ') & 077;
 			else
 			    b = 0;
-			if (s < send && *s >= ' ')
-			    c = (*s++ - ' ') & 077;
+			if (s < send && (unsigned char)*s >= ' ' && (unsigned char)*s < 'a')
+			    c = ((unsigned char)*s++ - ' ') & 077;
 			else
 			    c = 0;
-			if (s < send && *s >= ' ')
-			    d = (*s++ - ' ') & 077;
+			if (s < send && (unsigned char)*s >= ' ' && (unsigned char)*s < 'a')
+			    d = ((unsigned char)*s++ - ' ') & 077;
 			else
 			    d = 0;
 			hunk[0] = (char)(a << 2 | b >> 4);
@@ -1648,10 +1610,10 @@ pack_unpack(VALUE str, VALUE fmt)
 			ptr += mlen;
 			len -= mlen;
 		    }
-		    if (*s == '\r') s++;
-		    if (*s == '\n') s++;
-		    else if (s < send && (s+1 == send || s[1] == '\n'))
-			s += 2;	/* possible checksum byte */
+		    if (s < send && (unsigned char)*s != '\r' && *s != '\n')
+			s++;	/* possible checksum byte */
+		    if (s < send && *s == '\r') s++;
+		    if (s < send && *s == '\n') s++;
 		}
 
 		rb_str_set_len(buf, total);
@@ -1801,7 +1763,7 @@ pack_unpack(VALUE str, VALUE fmt)
 		    VALUE a;
 		    const VALUE *p, *pend;
 
-		    if (!(a = rb_str_associated(str))) {
+		    if (!(a = str_associated(str))) {
 			rb_raise(rb_eArgError, "no associated pointer");
 		    }
 		    p = RARRAY_CONST_PTR(a);
@@ -1810,7 +1772,7 @@ pack_unpack(VALUE str, VALUE fmt)
 			if (RB_TYPE_P(*p, T_STRING) && RSTRING_PTR(*p) == t) {
 			    if (len < RSTRING_LEN(*p)) {
 				tmp = rb_tainted_str_new(t, len);
-				rb_str_associate(tmp, a);
+				str_associate(tmp, a);
 			    }
 			    else {
 				tmp = *p;
@@ -1844,7 +1806,7 @@ pack_unpack(VALUE str, VALUE fmt)
 			VALUE a;
 			const VALUE *p, *pend;
 
-			if (!(a = rb_str_associated(str))) {
+			if (!(a = str_associated(str))) {
 			    rb_raise(rb_eArgError, "no associated pointer");
 			}
 			p = RARRAY_CONST_PTR(a);
@@ -1892,8 +1854,6 @@ pack_unpack(VALUE str, VALUE fmt)
     return ary;
 }
 
-#define BYTEWIDTH 8
-
 int
 rb_uv_to_utf8(char buf[6], unsigned long uv)
 {
@@ -2006,4 +1966,6 @@ Init_pack(void)
 {
     rb_define_method(rb_cArray, "pack", pack_pack, 1);
     rb_define_method(rb_cString, "unpack", pack_unpack, 1);
+
+    id_associated = rb_make_internal_id();
 }
