@@ -1,3 +1,4 @@
+# frozen_string_literal: false
 require 'test/unit'
 require 'date'
 
@@ -20,7 +21,6 @@ class TestDate < Test::Unit::TestCase
     assert_equal(7, Date::ABBR_DAYNAMES.size)
 
     assert(Date::MONTHNAMES.frozen?)
-    assert(!Date::MONTHNAMES[0].frozen?)
     assert(Date::MONTHNAMES[1].frozen?)
     assert(Date::DAYNAMES.frozen?)
     assert(Date::DAYNAMES[0].frozen?)
@@ -41,9 +41,7 @@ class TestDate < Test::Unit::TestCase
     assert_instance_of(DateSub, DateSub.today)
     assert_instance_of(DateTimeSub, DateTimeSub.now)
 
-#    assert_equal('#<DateSub: -4712-01-01 (-1/2,0,2299161)>', d.inspect)
     assert_equal('-4712-01-01', d.to_s)
-#    assert_equal('#<DateTimeSub: -4712-01-01T00:00:00+00:00 (-1/2,0,2299161)>', dt.inspect)
     assert_equal('-4712-01-01T00:00:00+00:00', dt.to_s)
 
     d2 = d + 1
@@ -131,6 +129,8 @@ class TestDate < Test::Unit::TestCase
     assert_equal(3, h.size)
     assert_equal(9, h[Date.new(1999,5,25)])
     assert_equal(9, h[DateTime.new(1999,5,25)])
+
+    assert_instance_of(String, Date.new(1999,5,25).hash.to_s)
   end
 
   def test_freeze
@@ -141,4 +141,13 @@ class TestDate < Test::Unit::TestCase
     assert_instance_of(String, d.to_s)
   end
 
+  def test_submillisecond_comparison
+    d1 = DateTime.new(2013, 12, 6, 0, 0, Rational(1, 10000))
+    d2 = DateTime.new(2013, 12, 6, 0, 0, Rational(2, 10000))
+    # d1 is 0.0001s earlier than d2
+    assert_equal(-1, d1 <=> d2)
+    assert_equal(0, d1 <=> d1)
+    assert_equal(1, d2 <=> d1)
+  end
+
 end
