@@ -1,37 +1,42 @@
-require 'test/unit'
-require File.join(File.expand_path(File.dirname(__FILE__)), 'gemutilities')
+# frozen_string_literal: true
+require 'rubygems/test_case'
 require 'rubygems/commands/dependency_command'
 
-class TestGemCommandsDependencyCommand < RubyGemTestCase
+class TestGemCommandsDependencyCommand < Gem::TestCase
 
   def setup
     super
 
     @cmd = Gem::Commands::DependencyCommand.new
     @cmd.options[:domain] = :local
-
-    util_setup_fake_fetcher
   end
 
   def test_execute
     quick_gem 'foo' do |gem|
       gem.add_dependency 'bar', '> 1'
+      gem.add_dependency 'baz', '> 1'
     end
 
-    Gem.source_index = nil
-
     @cmd.options[:args] = %w[foo]
 
     use_ui @ui do
       @cmd.execute
     end
 
-    assert_equal "Gem foo-2\n  bar (> 1, runtime)\n\n", @ui.output
+    assert_equal "Gem foo-2\n  bar (> 1)\n  baz (> 1)\n\n",
+                 @ui.output
     assert_equal '', @ui.error
   end
 
   def test_execute_no_args
-    Gem.source_index = nil
+    install_specs new_spec 'x', '2'
+
+    spec_fetcher do |fetcher|
+      fetcher.spec 'a', 1
+      fetcher.spec 'a', '2.a'
+      fetcher.spec 'dep_x', 1, 'x' => '>= 1'
+      fetcher.legacy_platform
+    end
 
     @cmd.options[:args] = []
 
@@ -42,16 +47,15 @@ class TestGemCommandsDependencyCommand < RubyGemTestCase
     expected = <<-EOF
 Gem a-1
 
-Gem a-2
+Gem a-2.a
 
-Gem a_evil-9
-
-Gem b-2
-
-Gem c-1.2
+Gem dep_x-1
+  x (>= 1)
 
 Gem pl-1-x86-linux
 
+Gem x-2
+
     EOF
 
     assert_equal expected, @ui.output
@@ -61,7 +65,7 @@ Gem pl-1-x86-linux
   def test_execute_no_match
     @cmd.options[:args] = %w[foo]
 
-    assert_raise MockGemUi::TermError do
+    assert_raises Gem::MockGemUi::TermError do
       use_ui @ui do
         @cmd.execute
       end
@@ -72,9 +76,11 @@ Gem pl-1-x86-linux
   end
 
   def test_execute_pipe_format
-    quick_gem 'foo' do |gem|
+    spec = util_spec 'foo' do |gem|
       gem.add_dependency 'bar', '> 1'
     end
+    install_specs util_spec 'bar', 2
+    install_specs spec
 
     @cmd.options[:args] = %w[foo]
     @cmd.options[:pipe_format] = true
@@ -88,7 +94,12 @@ Gem pl-1-x86-linux
   end
 
   def test_execute_regexp
-    Gem.source_index = nil
+    spec_fetcher do |fetcher|
+      fetcher.spec 'a',      1
+      fetcher.spec 'a',      '2.a'
+      fetcher.spec 'a_evil', 9
+      fetcher.spec 'b',      2
+    end
 
     @cmd.options[:args] = %w[/[ab]/]
 
@@ -99,7 +110,7 @@ Gem pl-1-x86-linux
     expected = <<-EOF
 Gem a-1
 
-Gem a-2
+Gem a-2.a
 
 Gem a_evil-9
 
@@ -112,6 +123,7 @@ Gem b-2
   end
 
   def test_execute_reverse
+    # FIX: this shouldn't need to write out, but fails if you switch it
     quick_gem 'foo' do |gem|
       gem.add_dependency 'bar', '> 1'
     end
@@ -120,8 +132,6 @@ Gem b-2
       gem.add_dependency 'foo'
     end
 
-    Gem.source_index = nil
-
     @cmd.options[:args] = %w[foo]
     @cmd.options[:reverse_dependencies] = true
 
@@ -131,9 +141,9 @@ Gem b-2
 
     expected = <<-EOF
 Gem foo-2
-  bar (> 1, runtime)
+  bar (> 1)
   Used by
-    baz-2 (foo (>= 0, runtime))
+    baz-2 (foo (>= 0))
 
     EOF
 
@@ -146,7 +156,7 @@ Gem foo-2
     @cmd.options[:reverse_dependencies] = true
     @cmd.options[:domain] = :remote
 
-    assert_raise MockGemUi::TermError do
+    assert_raises Gem::MockGemUi::TermError do
       use_ui @ui do
         @cmd.execute
       end
@@ -161,18 +171,12 @@ ERROR:  Only reverse dependencies for local gems are supported.
   end
 
   def test_execute_remote
-    foo = quick_gem 'foo' do |gem|
-      gem.add_dependency 'bar', '> 1'
+    install_specs new_spec 'bar', '2'
+
+    spec_fetcher do |fetcher|
+      fetcher.spec 'foo', 2, 'bar' => '> 1'
     end
 
-    @fetcher = Gem::FakeFetcher.new
-    Gem::RemoteFetcher.fetcher = @fetcher
-
-    util_setup_spec_fetcher foo
-
-    FileUtils.rm File.join(@gemhome, 'specifications',
-                           "#{foo.full_name}.gemspec")
-
     @cmd.options[:args] = %w[foo]
     @cmd.options[:domain] = :remote
 
@@ -180,47 +184,46 @@ ERROR:  Only reverse dependencies for local gems are supported.
       @cmd.execute
     end
 
-    assert_equal "Gem foo-2\n  bar (> 1, runtime)\n\n", @ui.output
+    assert_equal "Gem foo-2\n  bar (> 1)\n\n", @ui.output
     assert_equal '', @ui.error
   end
 
-  def test_execute_remote_legacy
-    foo = quick_gem 'foo' do |gem|
-      gem.add_dependency 'bar', '> 1'
-    end
-
+  def test_execute_remote_version
     @fetcher = Gem::FakeFetcher.new
     Gem::RemoteFetcher.fetcher = @fetcher
 
-    Gem::SpecFetcher.fetcher = nil
-    si = util_setup_source_info_cache foo
+    spec_fetcher do |fetcher|
+      fetcher.spec 'a', 1
+      fetcher.spec 'a', 2
+    end
 
-    @fetcher.data["#{@gem_repo}yaml"] = YAML.dump si
-    @fetcher.data["#{@gem_repo}Marshal.#{Gem.marshal_version}"] =
-      si.dump
-
-    @fetcher.data.delete "#{@gem_repo}latest_specs.#{Gem.marshal_version}.gz"
-
-    FileUtils.rm File.join(@gemhome, 'specifications',
-                           "#{foo.full_name}.gemspec")
-
-    @cmd.options[:args] = %w[foo]
+    @cmd.options[:args] = %w[a]
     @cmd.options[:domain] = :remote
+    @cmd.options[:version] = req '= 1'
 
     use_ui @ui do
       @cmd.execute
     end
 
-    assert_equal "Gem foo-2\n  bar (> 1, runtime)\n\n", @ui.output
+    assert_equal "Gem a-1\n\n", @ui.output
+    assert_equal '', @ui.error
+  end
 
-    expected = <<-EOF
-WARNING:  RubyGems 1.2+ index not found for:
-\t#{@gem_repo}
+  def test_execute_prerelease
+    spec_fetcher do |fetcher|
+      fetcher.spec 'a', '2.a'
+    end
 
-RubyGems will revert to legacy indexes degrading performance.
-    EOF
+    @cmd.options[:args] = %w[a]
+    @cmd.options[:domain] = :remote
+    @cmd.options[:prerelease] = true
 
-    assert_equal expected, @ui.error
+    use_ui @ui do
+      @cmd.execute
+    end
+
+    assert_equal "Gem a-2.a\n\n", @ui.output
+    assert_equal '', @ui.error
   end
 
 end
