@@ -2,7 +2,7 @@
 
   variable.c -
 
-  $Author: ngoto $
+  $Author$
   created at: Tue Apr 19 23:55:15 JST 1994
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
@@ -1605,7 +1605,7 @@ rb_ivar_count(VALUE obj)
     switch (BUILTIN_TYPE(obj)) {
       case T_OBJECT:
 	if ((tbl = ROBJECT_IV_INDEX_TBL(obj)) != 0) {
-	    st_index_t i, count, num = tbl->num_entries;
+	    st_index_t i, count, num = ROBJECT_NUMIV(obj);
 	    const VALUE *const ivptr = ROBJECT_IVPTR(obj);
 	    for (i = count = 0; i < num; ++i) {
 		if (ivptr[i] != Qundef) {
@@ -2020,6 +2020,17 @@ check_autoload_required(VALUE mod, ID id, const char **loadingpath)
     if (!RSTRING_PTR(file) || !*RSTRING_PTR(file)) {
 	rb_raise(rb_eArgError, "empty file name");
     }
+
+    /*
+     * if somebody else is autoloading, we MUST wait for them, since
+     * rb_provide_feature can provide a feature before autoload_const_set
+     * completes.  We must wait until autoload_const_set finishes in
+     * the other thread.
+     */
+    if (ele->state && ele->state->thread != rb_thread_current()) {
+	return load;
+    }
+
     loading = RSTRING_PTR(file);
     safe = rb_safe_level();
     rb_set_safe_level_force(0);
@@ -2127,7 +2138,7 @@ autoload_reset(VALUE arg)
 	    VALUE th = cur->thread;
 
 	    cur->thread = Qfalse;
-	    list_del(&cur->waitq.node);
+	    list_del_init(&cur->waitq.node); /* idempotent */
 
 	    /*
 	     * cur is stored on the stack of cur->waiting_th,
@@ -2140,6 +2151,34 @@ autoload_reset(VALUE arg)
     return 0;			/* ignored */
 }
 
+static VALUE
+autoload_sleep(VALUE arg)
+{
+    struct autoload_state *state = (struct autoload_state *)arg;
+
+    /*
+     * autoload_reset in other thread will resume us and remove us
+     * from the waitq list
+     */
+    do {
+	rb_thread_sleep_deadly();
+    } while (state->thread != Qfalse);
+
+    return Qfalse;
+}
+
+static VALUE
+autoload_sleep_done(VALUE arg)
+{
+    struct autoload_state *state = (struct autoload_state *)arg;
+
+    if (state->thread != Qfalse && rb_thread_to_be_killed(state->thread)) {
+	list_del_init(&state->waitq.node); /* idempotent */
+    }
+
+    return Qfalse;
+}
+
 VALUE
 rb_autoload_load(VALUE mod, ID id)
 {
@@ -2177,13 +2216,9 @@ rb_autoload_load(VALUE mod, ID id)
     }
     else {
 	list_add_tail(&ele->state->waitq.head, &state.waitq.node);
-	/*
-	 * autoload_reset in other thread will resume us and remove us
-	 * from the waitq list
-	 */
-	do {
-	    rb_thread_sleep_deadly();
-	} while (state.thread != Qfalse);
+
+	rb_ensure(autoload_sleep, (VALUE)&state,
+		autoload_sleep_done, (VALUE)&state);
     }
 
     /* autoload_data_i can be deleted by another thread while require */
@@ -2209,6 +2244,20 @@ rb_autoload_p(VALUE mod, ID id)
     return (ele = check_autoload_data(load)) ? ele->feature : Qnil;
 }
 
+void
+rb_const_warn_if_deprecated(const rb_const_entry_t *ce, VALUE klass, ID id)
+{
+    if (RB_CONST_DEPRECATED_P(ce)) {
+	if (klass == rb_cObject) {
+	    rb_warn("constant ::%"PRIsVALUE" is deprecated", QUOTE_ID(id));
+	}
+	else {
+	    rb_warn("constant %"PRIsVALUE"::%"PRIsVALUE" is deprecated",
+		    rb_class_name(klass), QUOTE_ID(id));
+	}
+    }
+}
+
 static VALUE
 rb_const_get_0(VALUE klass, ID id, int exclude, int recurse, int visibility)
 {
@@ -2224,17 +2273,9 @@ rb_const_get_0(VALUE klass, ID id, int exclude, int recurse, int visibility)
 	while ((ce = rb_const_lookup(tmp, id))) {
 	    if (visibility && RB_CONST_PRIVATE_P(ce)) {
 		rb_name_err_raise("private constant %2$s::%1$s referenced",
-				  klass, ID2SYM(id));
-	    }
-	    if (RB_CONST_DEPRECATED_P(ce)) {
-		if (klass == rb_cObject) {
-		    rb_warn("constant ::%"PRIsVALUE" is deprecated", QUOTE_ID(id));
-		}
-		else {
-		    rb_warn("constant %"PRIsVALUE"::%"PRIsVALUE" is deprecated",
-			    rb_class_name(klass), QUOTE_ID(id));
-		}
+				  tmp, ID2SYM(id));
 	    }
+	    rb_const_warn_if_deprecated(ce, tmp, id);
 	    value = ce->value;
 	    if (value == Qundef) {
 		if (am == tmp) break;
@@ -2370,7 +2411,9 @@ sv_i(st_data_t k, st_data_t v, st_data_t a)
 static int
 rb_local_constants_i(st_data_t const_name, st_data_t const_value, st_data_t ary)
 {
-    rb_ary_push((VALUE)ary, ID2SYM((ID)const_name));
+    if (rb_is_const_id(const_name) && !RB_CONST_PRIVATE_P((rb_const_entry_t *)const_value)) {
+	rb_ary_push((VALUE)ary, ID2SYM((ID)const_name));
+    }
     return ST_CONTINUE;
 }
 
@@ -2445,6 +2488,9 @@ rb_const_list(void *data)
  *  modules (example at start of section), unless the <i>inherit</i>
  *  parameter is set to <code>false</code>.
  *
+ *  The implementation makes no guarantees about the order in which the
+ *  constants are yielded.
+ *
  *    IO.constants.include?(:SYNC)        #=> true
  *    IO.constants(false).include?(:SYNC) #=> false
  *
