@@ -2,7 +2,7 @@
 
   vm.h -
 
-  $Author: usa $
+  $Author$
   created at: 04/01/01 16:56:59 JST
 
   Copyright (C) 2004-2007 Koichi Sasada
@@ -14,20 +14,9 @@
 
 typedef long OFFSET;
 typedef unsigned long lindex_t;
-typedef unsigned long dindex_t;
 typedef VALUE GENTRY;
 typedef rb_iseq_t *ISEQ;
 
-#ifdef  COLLECT_USAGE_ANALYSIS
-#define USAGE_ANALYSIS_INSN(insn)           vm_analysis_insn(insn)
-#define USAGE_ANALYSIS_OPERAND(insn, n, op) vm_analysis_operand((insn), (n), (VALUE)(op))
-#define USAGE_ANALYSIS_REGISTER(reg, s)     vm_analysis_register((reg), (s))
-#else
-#define USAGE_ANALYSIS_INSN(insn)		/* none */
-#define USAGE_ANALYSIS_OPERAND(insn, n, op)	/* none */
-#define USAGE_ANALYSIS_REGISTER(reg, s)		/* none */
-#endif
-
 #ifdef __GCC__
 /* TODO: machine dependent prefetch instruction */
 #define PREFETCH(pc)
@@ -38,7 +27,7 @@ typedef rb_iseq_t *ISEQ;
 #if VMDEBUG > 0
 #define debugs printf
 #define DEBUG_ENTER_INSN(insn) \
-  debug_print_pre(th, GET_CFP());
+    rb_vmdebug_debug_print_pre(th, GET_CFP(),GET_PC());
 
 #if OPT_STACK_CACHING
 #define SC_REGS() , reg_a, reg_b
@@ -47,7 +36,7 @@ typedef rb_iseq_t *ISEQ;
 #endif
 
 #define DEBUG_END_INSN() \
-  debug_print_post(th, GET_CFP() SC_REGS());
+  rb_vmdebug_debug_print_post(th, GET_CFP() SC_REGS());
 
 #else
 
@@ -60,7 +49,7 @@ typedef rb_iseq_t *ISEQ;
 /* #define throwdebug printf */
 
 /************************************************/
-#if   DISPATCH_XXX
+#if defined(DISPATCH_XXX)
 error !
 /************************************************/
 #elif OPT_CALL_THREADED_CODE
@@ -94,10 +83,10 @@ error !
   LABEL(insn): \
   INSN_ENTRY_SIG(insn); \
 
-/* dispather */
-#if __GNUC__ && (__i386__ || __x86_64__) && __GNUC__ == 3
+/* dispatcher */
+#if defined(__GNUC__) && (defined(__i386__) || defined(__x86_64__)) && __GNUC__ == 3
 #define DISPATCH_ARCH_DEPEND_WAY(addr) \
-  asm volatile("jmp *%0;\t# -- inseted by vm.h\t[length = 2]" : : "r" (addr))
+  __asm__ __volatile__("jmp *%0;\t# -- inserted by vm.h\t[length = 2]" : : "r" (addr))
 
 #else
 #define DISPATCH_ARCH_DEPEND_WAY(addr) \
@@ -116,7 +105,7 @@ error !
   ;
 
 #else
-/* token threade code */
+/* token threaded code */
 
 #define TC_DISPATCH(insn)  \
   DISPATCH_ARCH_DEPEND_WAY(insns_address_table[GET_CURRENT_INSN()]); \
@@ -129,7 +118,7 @@ error !
 
 #define END_INSN(insn)      \
   DEBUG_END_INSN();         \
-  TC_DISPATCH(insn);        \
+  TC_DISPATCH(insn);
 
 #define INSN_DISPATCH()     \
   TC_DISPATCH(__START__)    \
@@ -154,8 +143,8 @@ case BIN(insn):
 
 
 #define INSN_DISPATCH()         \
-  while(1){                     \
-    switch(GET_CURRENT_INSN()){
+  while (1) {			\
+    switch (GET_CURRENT_INSN()) {
 
 #define END_INSNS_DISPATCH()    \
 default:                        \
@@ -181,4 +170,13 @@ default:                        \
 
 #define SCREG(r) (reg_##r)
 
+#define VM_DEBUG_STACKOVERFLOW 0
+
+#if VM_DEBUG_STACKOVERFLOW
+#define CHECK_VM_STACK_OVERFLOW_FOR_INSN(cfp, margin) \
+    WHEN_VM_STACK_OVERFLOWED(cfp, (cfp)->sp, margin) vm_stack_overflow_for_insn()
+#else
+#define CHECK_VM_STACK_OVERFLOW_FOR_INSN(cfp, margin)
+#endif
+
 #endif /* RUBY_VM_EXEC_H */
